# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Sequence
import pytest

from ezdxf import edgeminer as em
from ezdxf.math import Vec3, rtree


class TestBasicRequirements:
    @pytest.fixture(params=[Vec3, em._Vertex], scope="class")
    def abc(self, request):
        cls = request.param
        a = cls((1, 2, 3))
        b = cls((1, 2, 3))
        c = cls((1, 2, 3))
        return a, b, c

    def test_vec3_requirements(self, abc):
        a, b, c = abc

        # Same locations have unique identities.
        # Maybe future me decides to return the same instance for same locations as
        # optimization, because Vec3 is immutable!
        assert a is not b
        assert a is not c
        assert b is not c

        # equality is EXACT same location and vice versa
        # floating point equality on bit-level!
        assert a == b
        assert a == c
        assert b == c

        # EXACT same location has same hash
        assert hash(a) == hash(b)
        assert hash(a) == hash(c)
        assert hash(b) == hash(c)

    def test_rtree_requirements(self, abc):
        rt = rtree.RTree(abc)
        assert len(rt) == 3, "expected multiple entries for the same location"
        result = list(rt.points_in_sphere(Vec3(1, 2, 3), radius=0.1))
        assert len(result) == 3, "expected multiple entries for the same location"
        assert set([id(v) for v in abc]) == set(
            [id(v) for v in result]
        ), "expected the identical instances"


class TestEdge:
    def test_init(self):
        edge = em.make_edge((0, 0), (1, 0))
        assert edge.start == Vec3(0, 0)
        assert edge.end == Vec3(1, 0)
        assert edge.length == 1.0
        assert edge.is_reverse is False
        assert edge.payload is None

    def test_edge_is_immutable(self):
        edge = em.make_edge((0, 0), (1, 0))
        with pytest.raises(AttributeError):
            edge.id = 0

    def test_identity(self):
        edge0 = em.make_edge((0, 0), (1, 0))
        edge1 = em.make_edge((0, 0), (1, 0))
        assert edge0 == edge0
        assert edge0 != edge1, "each edge should have an unique identity"
        assert edge0 == edge0.reversed(), "reversed copies represent the same edge"

    def test_reversed_copy(self):
        edge = em.make_edge((0, 0), (1, 0))
        clone = edge.reversed()
        assert edge == clone
        assert edge.id == clone.id
        assert edge.start == clone.end
        assert edge.end == clone.start
        assert edge.length == clone.length
        assert edge.is_reverse is (not clone.is_reverse)
        assert edge.payload is clone.payload

    def test_edge_can_be_used_in_sets(self):
        A = em.make_edge((0, 0), (1, 0))
        B = em.make_edge((1, 0), (1, 1))
        C = em.make_edge((1, 1), (0, 1))

        s1 = set([A, B])
        s2 = set([C, B])
        result = s1.intersection(s2)
        assert len(result) == 1
        assert B in result


class SimpleLoops:
    #   0   1   2
    # 1 +-C-+-G-+
    #   |   |   |
    #   D   B   F
    #   |   |   |
    # 0 +-A-+-E-+

    A = em.make_edge((0, 0), (1, 0), length=0.5, payload="A")
    B = em.make_edge((1, 0), (1, 1), payload="B")
    C = em.make_edge((1, 1), (0, 1), payload="C")
    D = em.make_edge((0, 1), (0, 0), payload="D")
    E = em.make_edge((1, 0), (2, 0), payload="E")
    F = em.make_edge((2, 0), (2, 1), payload="F")
    G = em.make_edge((2, 1), (1, 1), payload="G")


class TestEdgeDeposit(SimpleLoops):
    #   0   1   2
    # 1 +-C-+-G-+
    #   |   |   |
    #   D   B   F
    #   |   |   |
    # 0 +-A-+-E-+
    @pytest.fixture
    def edges(self):
        return [self.A, self.B, self.C, self.D, self.E, self.F, self.G]

    def test_get_degree_of_vertex(self, edges: list[em.Edge]):
        deposit = em.Deposit(edges)
        assert deposit.degree((0, 0)) == 2
        assert deposit.degree((1, 0)) == 3
        assert deposit.degree((-1, -1)) == 0, "not in deposit"

    def test_get_degree_of_vertices(self, edges: list[em.Edge]):
        deposit = em.Deposit(edges)
        assert deposit.degrees([(0, 0), (1, 0), (-1, -1)]) == (2, 3, 0)
        assert deposit.degrees([]) == ()

    def test_degree_counter(self, edges: list[em.Edge]):
        deposit = em.Deposit(edges)
        counter = deposit.degree_counter()
        assert counter[1] == 0
        assert counter[2] == 4
        assert counter[3] == 2
        assert deposit.max_degree == 3

    def test_unique_vertices(self, edges: list[em.Edge]):
        deposit = em.Deposit(edges)
        assert len(deposit.unique_vertices()) == 6

    def test_find_edges_linked_to_vertex_A_D(self):
        deposit = em.Deposit([self.A, self.B, self.C, self.D])
        edges = deposit.edges_linked_to(self.A.start)
        ids = set(e.id for e in edges)
        assert len(ids) == 2
        assert self.A.id in ids
        assert self.D.id in ids

    def test_find_edges_linked_to_vertex_A_G(self, edges: list[em.Edge]):
        deposit = em.Deposit(edges)
        linked_edges = deposit.edges_linked_to(self.B.end)
        ids = set(e.id for e in linked_edges)
        assert len(ids) == 3
        assert self.B.id in ids
        assert self.C.id in ids
        assert self.G.id in ids

    def test_find_nearest_edge(self):
        deposit = em.Deposit([self.A, self.B, self.C, self.D])
        edge = deposit.find_nearest_edge((0.5, 0.6))
        assert edge is self.C

    def test_build_network_A_D(self):
        deposit = em.Deposit([self.A, self.B, self.C, self.D])
        # network of all edges connected directly or indirectly to A
        network = deposit.find_network(self.A)
        assert len(network) == 4
        assert self.B in network
        assert self.C in network
        assert self.D in network

    def test_solitary_edge_is_a_network(self):
        deposit = em.Deposit([self.A, self.C])
        network = deposit.find_network(self.A)
        assert len(network) == 0

    def test_build_network_A_G(self, edges: list[em.Edge]):
        deposit = em.Deposit(edges)
        # network of all edges connected directly or indirectly to B
        network = deposit.find_network(self.B)
        assert len(network) == 7

    def test_build_all_networks(self, edges: list[em.Edge]):
        deposit = em.Deposit(edges)
        assert len(deposit.find_all_networks()) == 1

    def test_build_all_disconnected_networks(self):
        #   0   1   2   3
        # 1 +-C-+   +-G-+
        #   |   |   |   |
        #   D   B   H   F
        #   |   |   |   |
        # 0 +-A-+   +-E-+
        E = em.make_edge((2, 0), (3, 0), payload="E")
        F = em.make_edge((3, 0), (3, 1), payload="F")
        G = em.make_edge((3, 1), (2, 1), payload="G")
        H = em.make_edge((2, 1), (2, 0), payload="H")

        deposit = em.Deposit([self.A, self.B, self.C, self.D, E, F, G, H])
        assert len(deposit.find_all_networks()) == 2

    def test_build_all_networks_solitary_edges(self):
        deposit = em.Deposit([self.A, self.C, self.F])
        assert len(deposit.find_all_networks()) == 0, "a single edge is not a network"

    def test_find_loose_ends(self):
        deposit = em.Deposit([self.A, self.E, self.B, self.C, self.G])
        edges = set(deposit.find_leafs())
        assert len(edges) == 4
        assert self.B not in edges

    def test_single_edge_is_a_loose_ends(self):
        deposit = em.Deposit([self.A])
        edges = list(deposit.find_leafs())
        assert len(edges) == 1

    def test_loops_do_not_have_loose_ends(self):
        deposit = em.Deposit([self.A, self.B, self.C, self.D])
        edges = set(deposit.find_leafs())
        assert len(edges) == 0


class TestLoop:
    # +-C-+
    # |   |
    # D   B
    # |   |
    # +-A-+

    A = em.make_edge((0, 0), (1, 0))
    B = em.make_edge((1, 0), (1, 1))
    C = em.make_edge((1, 1), (0, 1))
    D = em.make_edge((0, 1), (0, 0))

    def test_loop_key(self):
        loop1 = (self.A, self.B, self.C)
        loop2 = (self.B, self.C, self.A)  # rotated edges, same loop

        assert em.loop_key(loop1) == em.loop_key(loop2)


def collect(chain: Sequence[em.Edge]):
    return ",".join(e.payload for e in chain)


def ordered_edges(edges: Sequence[em.Edge], reverse=False):
    """Returns the loop edges in key order."""
    edge_dict = {e.id: e for e in edges}
    return (edge_dict[eid] for eid in em.loop_key(edges, reverse=reverse))


def collect_ordered(chain: Sequence[em.Edge]) -> str:
    """Returns the payload as strings in key order.

    Key order:
        Loop starts with the edge with the smallest id.
    """
    if len(chain) == 0:
        return ""
    elif len(chain) == 1:
        return chain[0].payload  # type: ignore
    return ",".join([e.payload for e in ordered_edges(chain)])


class TestFindSequential:
    #   0   1   2
    # 1 +-E-+-D-+
    #   |       |
    #   F       C
    #   |       |
    # 0 +-A-+-B-+

    A = em.make_edge((0, 0), (1, 0), payload="A")
    B = em.make_edge((1, 0), (2, 0), payload="B")
    C = em.make_edge((2, 0), (2, 1), payload="C")
    D = em.make_edge((2, 1), (1, 1), payload="D")
    E = em.make_edge((1, 1), (0, 1), payload="E")
    F = em.make_edge((0, 1), (0, 0), payload="F")

    def test_is_forward_connected(self):
        assert em.is_forward_connected(self.A, self.B) is True
        assert em.is_forward_connected(self.A, self.F) is False

    def test_find_sequential(self):
        edges = [self.A, self.B, self.C, self.D, self.E, self.F]
        result = em.find_sequential_chain(edges)
        assert len(result) == 6
        assert result[0] is self.A
        assert result[-1] is self.F


class TestLoopFinderSimple(SimpleLoops):

    @pytest.fixture(scope="class")
    def netAD(self):
        return em.Deposit([self.A, self.B, self.C, self.D])

    @pytest.fixture(scope="class")
    def netAG(self):
        return em.Deposit([self.A, self.B, self.C, self.D, self.E, self.F, self.G])

    def test_find_any_loop(self, netAG):
        finder = em.LoopFinder(netAG)
        loop = finder.find_any_loop(start=self.A)
        assert len(loop) > 3

    def test_loop_A_B_C_D(self, netAD):
        finder = em.LoopFinder(netAD)
        finder.search(self.A)
        solutions = list(finder)
        assert len(solutions) == 1
        assert collect_ordered(solutions[0]) == "A,B,C,D"

    def test_loop_D_A_B_C(self, netAD):
        finder = em.LoopFinder(netAD)
        finder.search(self.D)
        solutions = list(finder)
        assert len(solutions) == 1
        assert collect_ordered(solutions[0]) == "A,B,C,D"

    def test_loop_A_to_D_unique_solutions(self, netAD):
        finder = em.LoopFinder(netAD)
        finder.search(self.A)
        # rotated edges, same loop
        finder.search(self.D)
        solutions = list(finder)
        assert len(solutions) == 1

    def test_loops_A_to_G(self, netAG):
        finder = em.LoopFinder(netAG, timeout=10)
        finder.search(self.A)
        solutions = list(finder)
        assert len(solutions) == 2
        expected = {"A,B,C,D", "A,E,F,G,C,D"}
        assert collect_ordered(solutions[0]) in expected
        assert collect_ordered(solutions[1]) in expected


def simple_loops() -> em.Deposit:
    #   0   1   2
    # 1 +-C-+-G-+
    #   |   |   |
    #   D   B   F
    #   |   |   |
    # 0 +-A-+-E-+
    return em.Deposit(
        [
            em.make_edge((0, 0), (1, 0), length=0.5, payload="A"),
            em.make_edge((1, 0), (1, 1), payload="B"),
            em.make_edge((1, 1), (0, 1), payload="C"),
            em.make_edge((0, 1), (0, 0), payload="D"),
            em.make_edge((1, 0), (2, 0), payload="E"),
            em.make_edge((2, 0), (2, 1), payload="F"),
            em.make_edge((2, 1), (1, 1), payload="G"),
        ]
    )


def complex_loops() -> Sequence[em.Edge]:
    #   0   1   2   3
    # 1 +-C-+-I-+-G-+
    #   |   |   |   |
    #   D   B   H   F
    #   |   |   |   |
    # 0 +-A-+-J-+-E-+

    return [
        em.make_edge((0, 0), (1, 0), payload="A"),
        em.make_edge((1, 0), (1, 1), payload="B"),
        em.make_edge((1, 1), (0, 1), payload="C"),
        em.make_edge((0, 1), (0, 0), payload="D"),
        em.make_edge((2, 0), (3, 0), payload="E"),
        em.make_edge((3, 0), (3, 1), payload="F"),
        em.make_edge((3, 1), (2, 1), payload="G"),
        em.make_edge((2, 1), (2, 0), payload="H"),
        em.make_edge((1, 1), (2, 1), payload="I"),
        em.make_edge((1, 0), (2, 0), payload="J"),
    ]


def test_find_all_sequential():
    #   0   1   2   3
    # 1 +-C-+-I-+-G-+
    #   |   |   |   |
    #   D   B   H   F
    #   |   |   |   |
    # 0 +-A-+-J-+-E-+
    edges = complex_loops()
    result = list(em.find_all_sequential_chains(edges))

    assert len(result) == 4
    assert collect_ordered(result[0]) == "A,B,C,D"
    assert collect_ordered(result[1]) == "E,F,G,H"
    assert collect_ordered(result[2]) == "I"
    assert collect_ordered(result[3]) == "J"


def grid() -> Sequence[em.Edge]:
    #   0   1   2
    # 2 +-F-+-E-+
    #   G   J   D
    # 1 +-K-+-L-+
    #   H   I   C
    # 0 +-A-+-B-+
    return [
        em.make_edge((0, 0), (1, 0), payload="A"),
        em.make_edge((1, 0), (2, 0), payload="B"),
        em.make_edge((2, 0), (2, 1), payload="C"),
        em.make_edge((2, 1), (2, 2), payload="D"),
        em.make_edge((2, 2), (1, 2), payload="E"),
        em.make_edge((1, 2), (0, 2), payload="F"),
        em.make_edge((0, 2), (0, 1), payload="G"),
        em.make_edge((0, 1), (0, 0), payload="H"),
        em.make_edge((1, 0), (1, 1), payload="I"),
        em.make_edge((1, 1), (1, 2), payload="J"),
        em.make_edge((0, 1), (1, 1), payload="K"),
        em.make_edge((1, 1), (2, 1), payload="L"),
    ]


def test_find_all_complex_loops():
    #   0   1   2
    # 2 +-F-+-E-+
    #   G   J   D
    # 1 +-K-+-L-+
    #   H   I   C
    # 0 +-A-+-B-+
    edges = grid()
    result = em.find_all_loops(em.Deposit(edges))
    assert len(result) == 13

    unique_loops = list(em.unique_chains(result))
    assert len(unique_loops) == 13


class TestAPIFunction:
    #   0   1   2
    # 1 +-C-+-G-+
    #   |   |   |
    #   D   B   F
    #   |   |   |
    # 0 +-A-+-E-+
    def test_find_all_loop(self):
        solutions = em.find_all_loops(simple_loops())
        assert len(solutions) == 3
        solution_strings = set(collect_ordered(s) for s in solutions)
        valid_solutions = {
            "A,B,C,D",  # forward
            "A,D,C,B",  # reverse
            "B,E,F,G",  # forward
            "B,G,F,E",  # reverse
            "A,E,F,G,C,D",  # forward
            "A,D,C,G,F,E",  # reverse
        }
        assert len(solution_strings.intersection(valid_solutions)) == 3

    def test_find_first_loop(self):
        solution = em.find_loop(simple_loops())
        assert len(solution) >= 4  # any loop is a valid solution

    def test_find_shortest_loop(self):
        solution = em.shortest_chain(em.find_all_loops(simple_loops()))
        assert len(solution) == 4
        assert collect_ordered(solution) == "A,B,C,D"

    def test_find_longest_loop(self):
        solution = em.longest_chain(em.find_all_loops(simple_loops()))
        assert len(solution) == 6
        assert collect_ordered(solution) == "A,E,F,G,C,D"


class TestFindAllDisconnectedLoops:
    #   0   1   2   3
    # 1 +-C-+   +-G-+
    #   |   |   |   |
    #   D   B   H   F
    #   |   |   |   |
    # 0 +-A-+   +-E-+

    A = em.make_edge((0, 0), (1, 0), payload="A")
    B = em.make_edge((1, 0), (1, 1), payload="B")
    C = em.make_edge((1, 1), (0, 1), payload="C")
    D = em.make_edge((0, 1), (0, 0), payload="D")
    E = em.make_edge((2, 0), (3, 0), payload="E")
    F = em.make_edge((3, 0), (3, 1), payload="F")
    G = em.make_edge((3, 1), (2, 1), payload="G")
    H = em.make_edge((2, 1), (2, 0), payload="H")

    def test_find_all_loops(self):
        solutions = em.find_all_loops(
            em.Deposit((self.A, self.B, self.C, self.D, self.E, self.F, self.G, self.H))
        )
        assert len(solutions) == 2
        solution_strings = [collect_ordered(s) for s in solutions]

        assert "A,B,C,D" in solution_strings
        assert "E,F,G,H" in solution_strings

    def test_find_all_shuffled_loops(self):
        solutions = em.find_all_loops(
            em.Deposit((self.H, self.B, self.F, self.D, self.E, self.C, self.G, self.A))
        )
        assert len(solutions) == 2
        solution_strings = [collect_ordered(s) for s in solutions]
        assert "A,B,C,D" in solution_strings
        assert "E,F,G,H" in solution_strings


class TestChainFinder:
    #    0   1   2   3   4   5
    #  2         G
    #  1         F
    #  0 +-A-+-B-+-C-+-D-+-E-+
    # -1         I
    # -2         J

    A = em.make_edge((0, 0), (1, 0), payload="A")
    B = em.make_edge((1, 0), (2, 0), payload="B")
    C = em.make_edge((2, 0), (3, 0), payload="C")
    D = em.make_edge((3, 0), (4, 0), payload="D")
    E = em.make_edge((4, 0), (5, 0), payload="E")

    F = em.make_edge((2, 0), (2, 1), payload="F")
    G = em.make_edge((2, 1), (2, 2), payload="G")
    I = em.make_edge((2, 0), (2, -1), payload="I")
    J = em.make_edge((2, -1), (2, -2), payload="J")

    def test_find_simple_chain(self):
        edges = [self.A, self.B, self.C, self.D, self.E]
        deposit = em.Deposit(edges)
        for edge in edges:
            result = em.find_simple_chain(deposit, edge)
            assert collect_ordered(result) == "A,B,C,D,E"

    def test_find_all_simple_chains(self):
        edges = [self.A, self.B, self.C, self.D, self.E, self.F, self.G, self.I, self.J]
        result = em.find_all_simple_chains(em.Deposit(edges))
        assert len(result) == 4

    def test_closed_loop(self):
        # 1 +-C-+
        #   |   |
        #   D   B
        #   |   |
        # 0 +-A-+
        A = em.make_edge((0, 0), (1, 0), payload="A")
        B = em.make_edge((1, 0), (1, 1), payload="B")
        C = em.make_edge((1, 1), (0, 1), payload="C")
        D = em.make_edge((0, 1), (0, 0), payload="D")
        deposit = em.Deposit([A, B, C, D])
        for edge in [A, B, C, D]:
            result = em.find_simple_chain(deposit, edge)
            assert collect_ordered(result) == "A,B,C,D"


class TestWrappingChains:
    #    0   1   2   3   4   5
    #  0 +-A-+-B-+-C-+-D-+-E-+
    A = em.make_edge((0, 0), (1, 0), payload="A")
    B = em.make_edge((1, 0), (2, 0), payload="B")
    C = em.make_edge((2, 0), (3, 0), payload="C")
    D = em.make_edge((3, 0), (4, 0), payload="D")
    E = em.make_edge((4, 0), (5, 0), payload="E")

    @pytest.fixture(scope="class")
    def edges(self):
        return (self.A, self.B, self.C, self.D, self.E)

    def test_wrap_chain(self, edges: list[em.Edge]):
        wrapped_chain = em.wrap_simple_chain(edges)
        wrapper = wrapped_chain.payload
        assert isinstance(wrapper, em.EdgeWrapper)
        assert wrapper.edges == edges

    def test_is_wrapped_chain(self, edges: list[em.Edge]):
        wrapped_chain = em.wrap_simple_chain(edges)
        assert em.is_wrapped_chain(wrapped_chain) is True
        assert em.is_wrapped_chain(self.A) is False

    def test_wrapping_empty_chain_raises_exception(self):
        with pytest.raises(ValueError):
            em.wrap_simple_chain([])

    def test_wrapping_single_edge_raises_exception(self):
        with pytest.raises(ValueError):
            em.wrap_simple_chain([self.A])

    def test_wrapping_unlinked_edges_raises_exception(self):
        with pytest.raises(ValueError):
            em.wrap_simple_chain([self.A, self.C])

    def test_wrapping_loop_raises_exception(self):
        with pytest.raises(ValueError):
            em.wrap_simple_chain([self.A, self.A.reversed()])

    def test_unwrap_chain(self, edges: list[em.Edge]):
        wrapped_chain = em.wrap_simple_chain(edges)
        chain = em.unwrap_simple_chain(wrapped_chain)
        assert len(chain) == 5
        assert chain == edges

    def test_unwrap_reversed_chain(self, edges: list[em.Edge]):
        wrapped_chain = em.wrap_simple_chain(edges)
        reversed_edge = wrapped_chain.reversed()
        chain = em.unwrap_simple_chain(reversed_edge)
        assert len(chain) == 5
        assert chain[0].start == reversed_edge.start
        assert chain[-1].end == reversed_edge.end

        assert chain[0] == edges[-1]
        assert chain[0].is_reverse is not edges[-1].is_reverse
        assert chain[-1] == edges[0]
        assert chain[-1].is_reverse is not edges[0].is_reverse

    def test_unwrapping_single_edge(self):
        edges = em.unwrap_simple_chain(self.A)
        assert len(edges) == 1
        assert edges[0] == self.A

    def test_flatten_nested_edges(self):
        de = em.wrap_simple_chain([self.D, self.E])
        ab = em.wrap_simple_chain([self.A, self.B])
        cde = em.wrap_simple_chain([self.C, de])
        abcde = em.wrap_simple_chain([ab, cde])
        assert collect_ordered(list(em.flatten(abcde))) == "A,B,C,D,E"

    def test_flatten_empty_sequence(self):
        assert len(list(em.flatten([]))) == 0


class TestOpenChainFinder:

    def test_find_all_open_chains(self):
        #   0   1   2   3   4
        # 3 +---+---+-E-+-F-+
        #   |   |   D   |   |
        # 2 +-A-+-B-+-C-+---+
        #   |   G   |   |   |
        # 1 +---+-H-+---+---+
        #   |   |   I   |   |
        # 0 +---+---+---+---+
        # all end to end connections:
        # - 3 ABC or CBA
        # - 4 AGHI
        # - 4 C..F
        # - 5 A...F
        # - 5 C...I
        # - 7 I.....F

        A = em.make_edge((0, 2), (1, 2), payload="A")
        B = em.make_edge((1, 2), (2, 2), payload="B")
        C = em.make_edge((2, 2), (3, 2), payload="C")
        D = em.make_edge((2, 2), (2, 3), payload="D")
        E = em.make_edge((2, 3), (3, 3), payload="E")
        F = em.make_edge((3, 3), (4, 3), payload="F")
        G = em.make_edge((1, 2), (1, 1), payload="G")
        H = em.make_edge((1, 1), (2, 1), payload="H")
        I = em.make_edge((2, 1), (2, 0), payload="I")
        edges = (H, C, E, D, B, G, F, A, I)
        combinations = em.find_all_open_chains(em.Deposit(edges))
        assert len(combinations) == 6
        assert len(combinations[0]) == 3
        assert collect(combinations[0]) in ("A,B,C", "C,B,A")
        assert len(combinations[-1]) == 7
        assert collect(combinations[-1]) in ("I,H,G,B,D,E,F", "F,E,D,B,G,H,I")

    def test_does_not_detect_closed_loops(self):
        # 1 +-C-+
        #   |   |
        #   D   B
        #   |   |
        # 0 +-A-+
        A = em.make_edge((0, 0), (1, 0))
        B = em.make_edge((1, 0), (1, 1))
        C = em.make_edge((1, 1), (0, 1))
        D = em.make_edge((0, 1), (0, 0))
        deposit = em.Deposit([A, B, C, D])
        assert len(em.find_all_open_chains(deposit)) == 0

    def test_not_does_detect_indirect_loops(self):
        # 1 +-C-+
        #   |   |
        #   D   B
        #   |   |
        # 0 +-A-+-E-+
        A = em.make_edge((0, 0), (1, 0), payload="A")
        B = em.make_edge((1, 0), (1, 1), payload="B")
        C = em.make_edge((1, 1), (0, 1), payload="C")
        D = em.make_edge((0, 1), (0, 0), payload="D")
        E = em.make_edge((1, 0), (2, 0), payload="E")
        deposit = em.Deposit([A, B, C, D, E])
        result = set(collect(s) for s in em.find_all_open_chains(deposit))
        assert len(result) == 0


class TestFindLoopByEdge:
    #   0   1   2
    # 2 +-F-+-E-+
    #   G   J   D
    # 1 +-K-+-L-+
    #   H   I   C
    # 0 +-A-+-B-+
    edges = grid()

    def edge(self, payload: str):
        for edge in self.edges:
            if edge.payload == payload:
                return edge
        raise ValueError(f"edge {payload} does not exist")

    def test_search_continuation_clockwise(self):
        loop = em.find_loop_by_edge(
            em.Deposit(self.edges), self.edge("A"), clockwise=True
        )
        assert len(loop) == 4
        assert collect(loop) == "A,I,K,H"

    def test_search_continuation_counter_clockwise(self):
        loop = em.find_loop_by_edge(
            em.Deposit(self.edges), self.edge("A"), clockwise=False
        )
        assert len(loop) == 8
        assert collect(loop) == "A,B,C,D,E,F,G,H"


def test_filter_coincident_edges():
    edges = list(grid())
    edges.extend(grid())  # 2x the same edges
    assert len(em.filter_coincident_edges(em.Deposit(edges))) == 12


class TestFilterCloseVertices:
    def test_coincident_vertices(self):
        vertices = Vec3.list([(0, 0), (0, 0), (1, 1), (1, 1)])
        rt = rtree.RTree(vertices)
        result = em.filter_close_vertices(rt, gap_tol=1e-9)
        # You don't know which vertices were removed!
        assert len(result) == 2

    def test_chain_of_close_vertices(self):
        vertices = Vec3.list([(0, 0), (1, 0), (2, 0), (3, 0)])
        rt = rtree.RTree(vertices)
        result = em.filter_close_vertices(rt, gap_tol=1)
        # You don't know which vertices were removed!
        assert len(result) == 2

    def test_grid_of_close_vertices(self):
        # fmt: off
        vertices = Vec3.list([
            (0, 0), (1, 0), (2, 0), (3, 0),
            (0, 1), (1, 1), (2, 1), (3, 1),
            (0, 2), (1, 2), (2, 2), (3, 2),
            (0, 3), (1, 3), (2, 3), (3, 3)
        ])
        # fmt: on
        rt = rtree.RTree(vertices)
        result = em.filter_close_vertices(rt, gap_tol=1)
        # You don't know which vertices were removed!
        assert len(result) == 8


class TestSortEdgesByAngle:
    #   0   1   2
    # 2 +---+---+
    #   |\  |  /|
    #   | D C B |
    #   |  \|/  |
    # 1 +-E-+-A-+
    #   |  /|\  |
    #   | F G H |
    #   |/  |  \|
    # 0 +---+---+
    A = em.make_edge((1, 1), (2, 1), payload="A")
    B = em.make_edge((1, 1), (2, 2), payload="B")
    C = em.make_edge((1, 1), (1, 2), payload="C")
    D = em.make_edge((1, 1), (0, 2), payload="D")
    E = em.make_edge((1, 1), (0, 1), payload="E")
    F = em.make_edge((1, 1), (0, 0), payload="F")
    G = em.make_edge((1, 1), (1, 0), payload="G")
    H = em.make_edge((1, 1), (2, 0), payload="H")

    def test_edges_B_C_base_A(self):
        edges = [self.C, self.B]
        base = self.A.reversed()
        result = em.sort_edges_to_base(edges, base)
        assert result[0] is self.B
        assert result[1] is self.C

    def test_edges_G_H_base_A(self):
        edges = [self.G, self.H]
        base = self.A.reversed()
        result = em.sort_edges_to_base(edges, base)
        assert result[0] is self.G
        assert result[1] is self.H

    def test_edges_B_H_base_A(self):
        edges = [self.H, self.B]
        base = self.A.reversed()
        result = em.sort_edges_to_base(edges, base)
        assert result[0] is self.B
        assert result[1] is self.H

    def test_edges_A_B_base_C(self):
        edges = [self.A, self.B]
        base = self.C.reversed()
        result = em.sort_edges_to_base(edges, base)
        assert result[0] is self.A
        assert result[1] is self.B

    def test_edges_D_E_base_C(self):
        edges = [self.D, self.E]
        base = self.C.reversed()
        result = em.sort_edges_to_base(edges, base)
        assert result[0] is self.D
        assert result[1] is self.E

    def test_edges_B_D_base_C(self):
        edges = [self.B, self.D]
        base = self.C.reversed()
        result = em.sort_edges_to_base(edges, base)
        assert result[0] is self.D
        assert result[1] is self.B


class TestSubtractEdges:
    def test_subtract_nothing(self):
        edges = list(grid())
        result = em.subtract_edges(edges, [])
        assert len(result) == len(edges)

    def test_subtract_from_nothing(self):
        edges = list(grid())
        result = em.subtract_edges([], edges)
        assert len(result) == 0

    def test_subtract_one_edge(self):
        edges = list(grid())
        first = edges[0]
        result = em.subtract_edges(edges, [first])
        assert len(result) == len(edges) - 1
        assert first not in result

    def test_subtract_two_edges(self):
        edges = list(grid())
        two = edges[:2]
        result = em.subtract_edges(edges, two)
        assert len(result) == len(edges) - 2
        assert two[0] not in result
        assert two[1] not in result


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2010-2022, Manfred Moitzi
# License: MIT License
"""Dimension lines as composite entities build up by DXF primitives.

This add-on exist just for an easy transition from `dxfwrite` to `ezdxf`.

Classes
-------

- LinearDimension
- AngularDimension
- ArcDimension
- RadiusDimension
- DimStyle

This code was written long before I had any understanding of the DIMENSION
entity and therefore, the classes have completely different implementations and
styling features than the dimensions based on the DIMENSION entity .

.. warning::

    Try to not use these classes beside porting `dxfwrite` code to `ezdxf`
    and even for that case the usage of the regular DIMENSION entity is to
    prefer because this module will not get much maintenance and may be removed
    in the future.

"""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Iterable, Optional
from math import radians, degrees, pi
from abc import abstractmethod
from ezdxf.enums import TextEntityAlignment
from ezdxf.math import Vec3, Vec2, distance, lerp, ConstructionRay, UVec

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.eztypes import GenericLayoutType

DIMENSIONS_MIN_DISTANCE = 0.05
DIMENSIONS_FLOATINGPOINT = "."

ANGLE_DEG = 180.0 / pi
ANGLE_GRAD = 200.0 / pi
ANGLE_RAD = 1.0


class DimStyle(dict):
    """DimStyle parameter struct, a dumb object just to store values
    """

    default_values = [
        # tick block name, use setup to generate default blocks <dimblk> <dimblk1> <dimblk2>
        ("tick", "DIMTICK_ARCH"),
        # scale factor for ticks-block <dimtsz> <dimasz>
        ("tickfactor", 1.0),
        # tick2x means tick is drawn only for one side, insert tick a second
        # time rotated about 180 degree, but only one time at the dimension line
        # ends, this is useful for arrow-like ticks. hint: set dimlineext to 0. <none>
        ("tick2x", False),
        # dimension value scale factor, value = drawing-units * scale <dimlfac>
        ("scale", 100.0),
        # round dimension value to roundval fractional digits <dimdec>
        ("roundval", 0),
        # round dimension value to half units, round 0.4, 0.6 to 0.5 <dimrnd>
        ("roundhalf", False),
        # dimension value text color <dimclrt>
        ("textcolor", 7),
        # dimension value text height <dimtxt>
        ("height", 0.5),
        # dimension text prefix and suffix like 'x=' ... ' cm' <dimpost>
        ("prefix", ""),
        ("suffix", ""),
        # dimension value text style <dimtxsty>
        ("style", "OpenSansCondensed-Light"),
        # default layer for whole dimension object
        ("layer", "DIMENSIONS"),
        # dimension line color index (0 from layer) <dimclrd>
        ("dimlinecolor", 7),
        # dimension line extensions (in dimline direction, left and right) <dimdle>
        ("dimlineext", 0.3),
        # draw dimension value text `textabove` drawing-units above the
        # dimension line <dimgap>
        ("textabove", 0.2),
        # switch extension line False=off, True=on <dimse1> <dimse2>
        ("dimextline", True),
        # dimension extension line color index (0 from layer) <dimclre>
        ("dimextlinecolor", 5),
        # gap between measure target point and end of extension line <dimexo>
        ("dimextlinegap", 0.3),
    ]

    def __init__(self, name: str, **kwargs):
        super().__init__(DimStyle.default_values)
        # dimstyle name
        self["name"] = name
        self.update(kwargs)

    def __getattr__(self, attr: str) -> Any:
        return self[attr]

    def __setattr__(self, attr: str, value: Any) -> None:
        self[attr] = value


class DimStyles:
    """
    DimStyle container

    """

    def __init__(self) -> None:
        self._styles: dict[str, DimStyle] = {}
        self.default = DimStyle("Default")

        self.new(
            "angle.deg",
            scale=ANGLE_DEG,
            suffix=str("°"),
            roundval=0,
            tick="DIMTICK_RADIUS",
            tick2x=True,
            dimlineext=0.0,
            dimextline=False,
        )
        self.new(
            "angle.grad",
            scale=ANGLE_GRAD,
            suffix="gon",
            roundval=0,
            tick="DIMTICK_RADIUS",
            tick2x=True,
            dimlineext=0.0,
            dimextline=False,
        )
        self.new(
            "angle.rad",
            scale=ANGLE_RAD,
            suffix="rad",
            roundval=3,
            tick="DIMTICK_RADIUS",
            tick2x=True,
            dimlineext=0.0,
            dimextline=False,
        )

    def get(self, name: str) -> DimStyle:
        """
        Get DimStyle() object by name.
        """
        return self._styles.get(name, self.default)

    def new(self, name: str, **kwargs) -> DimStyle:
        """
        Create a new dimstyle
        """
        style = DimStyle(name, **kwargs)
        self._styles[name] = style
        return style

    @staticmethod
    def setup(drawing: "Drawing"):
        """
        Insert necessary definitions into drawing:

            ticks: DIMTICK_ARCH, DIMTICK_DOT, DIMTICK_ARROW
        """
        # default pen assignment:
        # 1 : 1.40mm - red
        # 2 : 0.35mm - yellow
        # 3 : 0.70mm - green
        # 4 : 0.50mm - cyan
        # 5 : 0.13mm - blue
        # 6 : 1.00mm - magenta
        # 7 : 0.25mm - white/black
        # 8, 9 : 2.00mm
        # >=10 : 1.40mm

        dimcolor = {
            "color": dimstyles.default.dimextlinecolor,
            "layer": "BYBLOCK",
        }
        color4 = {"color": 4, "layer": "BYBLOCK"}
        color7 = {"color": 7, "layer": "BYBLOCK"}

        block = drawing.blocks.new("DIMTICK_ARCH")
        block.add_line(start=(0.0, +0.5), end=(0.0, -0.5), dxfattribs=dimcolor)
        block.add_line(start=(-0.2, -0.2), end=(0.2, +0.2), dxfattribs=color4)

        block = drawing.blocks.new("DIMTICK_DOT")
        block.add_line(start=(0.0, 0.5), end=(0.0, -0.5), dxfattribs=dimcolor)
        block.add_circle(center=(0, 0), radius=0.1, dxfattribs=color4)

        block = drawing.blocks.new("DIMTICK_ARROW")

        block.add_line(start=(0.0, 0.5), end=(0.0, -0.50), dxfattribs=dimcolor)
        block.add_solid([(0, 0), (0.3, 0.05), (0.3, -0.05)], dxfattribs=color7)

        block = drawing.blocks.new("DIMTICK_RADIUS")
        block.add_solid(
            [(0, 0), (0.3, 0.05), (0.25, 0.0), (0.3, -0.05)], dxfattribs=color7
        )


dimstyles = DimStyles()  # use this factory to create new dimstyles


class _DimensionBase:
    """
    Abstract base class for dimension lines.

    """

    def __init__(self, dimstyle: str, layer: str, roundval: int):
        self.dimstyle = dimstyles.get(dimstyle)
        self.layer = layer
        self.roundval = roundval

    def prop(self, property_name: str) -> Any:
        """
        Get dimension line properties by `property_name` with the possibility to override several properties.
        """
        if property_name == "layer":
            return self.layer if self.layer is not None else self.dimstyle.layer
        elif property_name == "roundval":
            return (
                self.roundval
                if self.roundval is not None
                else self.dimstyle.roundval
            )
        else:  # pass through self.dimstyle object DimStyle()
            return self.dimstyle[property_name]

    def format_dimtext(self, dimvalue: float) -> str:
        """
        Format the dimension text.
        """
        dimtextfmt = "%." + str(self.prop("roundval")) + "f"
        dimtext = dimtextfmt % dimvalue
        if DIMENSIONS_FLOATINGPOINT in dimtext:
            # remove successional zeros
            dimtext.rstrip("0")
            # remove floating point as last char
            dimtext.rstrip(DIMENSIONS_FLOATINGPOINT)
        return self.prop("prefix") + dimtext + self.prop("suffix")

    @abstractmethod
    def render(self, layout: "GenericLayoutType"):
        pass


class LinearDimension(_DimensionBase):
    """
    Simple straight dimension line with two or more measure points, build with basic DXF entities. This is NOT a dxf
    dimension entity. And This is a 2D element, so all z-values will be ignored!

    """

    def __init__(
        self,
        pos: UVec,
        measure_points: Iterable[UVec],
        angle: float = 0.0,
        dimstyle: str = "Default",
        layer: Optional[str] = None,
        roundval: Optional[int] = None,
    ):
        """
        LinearDimension Constructor.

        Args:
            pos: location as (x, y) tuple of dimension line, line goes through this point
            measure_points: list of points as (x, y) tuples to dimension (two or more)
            angle: angle (in degree) of dimension line
            dimstyle: dimstyle name, 'Default' - style is the default value
            layer: dimension line layer, override the default value of dimstyle
            roundval: count of decimal places

        """
        super().__init__(dimstyle, layer, roundval)  # type: ignore
        self.angle = angle
        self.measure_points = list(measure_points)
        self.text_override = [""] * self.section_count
        self.dimlinepos = Vec3(pos)
        self.layout = None

    def set_text(self, section: int, text: str) -> None:
        """
        Set and override the text of the dimension text for the given dimension line section.
        """
        self.text_override[section] = text

    def _setup(self) -> None:
        """
        Calc setup values and determines the point order of the dimension line points.
        """
        self.measure_points = [Vec3(point) for point in self.measure_points]
        dimlineray = ConstructionRay(self.dimlinepos, angle=radians(self.angle))
        self.dimline_points = [
            self._get_point_on_dimline(point, dimlineray)
            for point in self.measure_points
        ]
        self.point_order = self._indices_of_sorted_points(self.dimline_points)
        self._build_vectors()

    def _get_dimline_point(self, index: int) -> UVec:
        """
        Get point on the dimension line, index runs left to right.
        """
        return self.dimline_points[self.point_order[index]]

    def _get_section_points(self, section: int) -> tuple[Vec3, Vec3]:
        """
        Get start and end point on the dimension line of dimension section.
        """
        return self._get_dimline_point(section), self._get_dimline_point(
            section + 1
        )

    def _get_dimline_bounds(self) -> tuple[Vec3, Vec3]:
        """
        Get the first and the last point of dimension line.
        """
        return self._get_dimline_point(0), self._get_dimline_point(-1)

    @property
    def section_count(self) -> int:
        """count of dimline sections"""
        return len(self.measure_points) - 1

    @property
    def point_count(self) -> int:
        """count of dimline points"""
        return len(self.measure_points)

    def render(self, layout: "GenericLayoutType") -> None:
        """build dimension line object with basic dxf entities"""
        self._setup()
        self._draw_dimline(layout)
        if self.prop("dimextline"):
            self._draw_extension_lines(layout)
        self._draw_text(layout)
        self._draw_ticks(layout)

    @staticmethod
    def _indices_of_sorted_points(points: Iterable[UVec]) -> list[int]:
        """get indices of points, for points sorted by x, y values"""
        indexed_points = [(point, idx) for idx, point in enumerate(points)]
        indexed_points.sort()
        return [idx for point, idx in indexed_points]

    def _build_vectors(self) -> None:
        """build unit vectors, parallel and normal to dimension line"""
        point1, point2 = self._get_dimline_bounds()
        self.parallel_vector = (Vec3(point2) - Vec3(point1)).normalize()
        self.normal_vector = self.parallel_vector.orthogonal()

    @staticmethod
    def _get_point_on_dimline(point: UVec, dimray: ConstructionRay) -> Vec3:
        """get the measure target point projection on the dimension line"""
        return dimray.intersect(dimray.orthogonal(point))

    def _draw_dimline(self, layout: "GenericLayoutType") -> None:
        """build dimension line entity"""
        start_point, end_point = self._get_dimline_bounds()

        dimlineext = self.prop("dimlineext")
        if dimlineext > 0:
            start_point = start_point - (self.parallel_vector * dimlineext)
            end_point = end_point + (self.parallel_vector * dimlineext)

        attribs = {
            "color": self.prop("dimlinecolor"),
            "layer": self.prop("layer"),
        }
        layout.add_line(
            start=start_point,
            end=end_point,
            dxfattribs=attribs,
        )

    def _draw_extension_lines(self, layout: "GenericLayoutType") -> None:
        """build the extension lines entities"""
        dimextlinegap = self.prop("dimextlinegap")
        attribs = {
            "color": self.prop("dimlinecolor"),
            "layer": self.prop("layer"),
        }

        for dimline_point, target_point in zip(
            self.dimline_points, self.measure_points
        ):
            if distance(dimline_point, target_point) > max(
                dimextlinegap, DIMENSIONS_MIN_DISTANCE
            ):
                direction_vector = (target_point - dimline_point).normalize()
                target_point = target_point - (direction_vector * dimextlinegap)
                layout.add_line(
                    start=dimline_point,
                    end=target_point,
                    dxfattribs=attribs,
                )

    def _draw_text(self, layout: "GenericLayoutType") -> None:
        """build the dimension value text entity"""
        attribs = {
            "height": self.prop("height"),
            "color": self.prop("textcolor"),
            "layer": self.prop("layer"),
            "rotation": self.angle,
            "style": self.prop("style"),
        }
        for section in range(self.section_count):
            dimvalue_text = self._get_dimvalue_text(section)
            insert_point = self._get_text_insert_point(section)
            layout.add_text(
                text=dimvalue_text,
                dxfattribs=attribs,
            ).set_placement(
                insert_point, align=TextEntityAlignment.MIDDLE_CENTER
            )

    def _get_dimvalue_text(self, section: int) -> str:
        """get the dimension value as text, distance from point1 to point2"""
        override = self.text_override[section]
        if len(override):
            return override
        point1, point2 = self._get_section_points(section)

        dimvalue = distance(point1, point2) * self.prop("scale")
        return self.format_dimtext(dimvalue)

    def _get_text_insert_point(self, section: int) -> Vec3:
        """get the dimension value text insert point"""
        point1, point2 = self._get_section_points(section)
        dist = self.prop("height") / 2.0 + self.prop("textabove")
        return lerp(point1, point2) + (self.normal_vector * dist)

    def _draw_ticks(self, layout: "GenericLayoutType") -> None:
        """insert the dimension line ticks, (markers on the dimension line)"""
        attribs = {
            "xscale": self.prop("tickfactor"),
            "yscale": self.prop("tickfactor"),
            "layer": self.prop("layer"),
        }

        def add_tick(index: int, rotate: bool = False) -> None:
            """build the insert-entity for the tick block"""
            attribs["rotation"] = self.angle + (180.0 if rotate else 0.0)
            layout.add_blockref(
                insert=self._get_dimline_point(index),
                name=self.prop("tick"),
                dxfattribs=attribs,
            )

        if self.prop("tick2x"):
            for index in range(0, self.point_count - 1):
                add_tick(index, False)
            for index in range(1, self.point_count):
                add_tick(index, True)
        else:
            for index in range(self.point_count):
                add_tick(index, False)


class AngularDimension(_DimensionBase):
    """
    Draw an angle dimensioning line at dimline pos from start to end, dimension text is the angle build of the three
    points start-center-end.

    """

    DEG = ANGLE_DEG
    GRAD = ANGLE_GRAD
    RAD = ANGLE_RAD

    def __init__(
        self,
        pos: UVec,
        center: UVec,
        start: UVec,
        end: UVec,
        dimstyle: str = "angle.deg",
        layer: Optional[str] = None,
        roundval: Optional[int] = None,
    ):
        """
        AngularDimension constructor.

        Args:
            pos: location as (x, y) tuple of dimension line, line goes through this point
            center: center point as (x, y) tuple of angle
            start: line from center to start is the first side of the angle
            end: line from center to end is the second side of the angle
            dimstyle: dimstyle name, 'Default' - style is the default value
            layer: dimension line layer, override the default value of dimstyle
            roundval: count of decimal places

        """
        super().__init__(dimstyle, layer, roundval)  # type: ignore
        self.dimlinepos = Vec3(pos)
        self.center = Vec3(center)
        self.start = Vec3(start)
        self.end = Vec3(end)

    def _setup(self) -> None:
        """setup calculation values"""
        self.pos_radius = distance(self.center, self.dimlinepos)  # type: float
        self.radius = distance(self.center, self.start)  # type: float
        self.start_vector = (self.start - self.center).normalize()  # type: Vec3
        self.end_vector = (self.end - self.center).normalize()  # type: Vec3
        self.start_angle = self.start_vector.angle  # type: float
        self.end_angle = self.end_vector.angle  # type: float

    def render(self, layout: "GenericLayoutType") -> None:
        """build dimension line object with basic dxf entities"""

        self._setup()
        self._draw_dimension_line(layout)
        if self.prop("dimextline"):
            self._draw_extension_lines(layout)
        self._draw_dimension_text(layout)
        self._draw_ticks(layout)

    def _draw_dimension_line(self, layout: "GenericLayoutType") -> None:
        """draw the dimension line from start- to endangle."""
        layout.add_arc(
            radius=self.pos_radius,
            center=self.center,
            start_angle=degrees(self.start_angle),
            end_angle=degrees(self.end_angle),
            dxfattribs={
                "layer": self.prop("layer"),
                "color": self.prop("dimlinecolor"),
            },
        )

    def _draw_extension_lines(self, layout: "GenericLayoutType") -> None:
        """build the extension lines entities"""
        for vector in [self.start_vector, self.end_vector]:
            layout.add_line(
                start=self._get_extline_start(vector),
                end=self._get_extline_end(vector),
                dxfattribs={
                    "layer": self.prop("layer"),
                    "color": self.prop("dimextlinecolor"),
                },
            )

    def _get_extline_start(self, vector: Vec3) -> Vec3:
        return self.center + (vector * self.prop("dimextlinegap"))

    def _get_extline_end(self, vector: Vec3) -> Vec3:
        return self.center + (vector * self.pos_radius)

    def _draw_dimension_text(self, layout: "GenericLayoutType") -> None:
        attribs = {
            "height": self.prop("height"),
            "rotation": degrees(
                (self.start_angle + self.end_angle) / 2 - pi / 2.0
            ),
            "layer": self.prop("layer"),
            "style": self.prop("style"),
            "color": self.prop("textcolor"),
        }
        layout.add_text(
            text=self._get_dimtext(),
            dxfattribs=attribs,
        ).set_placement(
            self._get_text_insert_point(),
            align=TextEntityAlignment.MIDDLE_CENTER,
        )

    def _get_text_insert_point(self) -> Vec3:
        midvector = ((self.start_vector + self.end_vector) / 2.0).normalize()
        length = (
            self.pos_radius + self.prop("textabove") + self.prop("height") / 2.0
        )
        return self.center + (midvector * length)

    def _draw_ticks(self, layout: "GenericLayoutType") -> None:
        attribs = {
            "xscale": self.prop("tickfactor"),
            "yscale": self.prop("tickfactor"),
            "layer": self.prop("layer"),
        }
        for vector, mirror in [
            (self.start_vector, False),
            (self.end_vector, self.prop("tick2x")),
        ]:
            insert_point = self.center + (vector * self.pos_radius)
            rotation = vector.angle + pi / 2.0
            attribs["rotation"] = degrees(rotation + (pi if mirror else 0.0))
            layout.add_blockref(
                insert=insert_point,
                name=self.prop("tick"),
                dxfattribs=attribs,
            )

    def _get_dimtext(self) -> str:
        # set scale = ANGLE_DEG for degrees (circle = 360 deg)
        # set scale = ANGLE_GRAD for grad(circle = 400 grad)
        # set scale = ANGLE_RAD for rad(circle = 2*pi)
        angle = (self.end_angle - self.start_angle) * self.prop("scale")
        return self.format_dimtext(angle)


class ArcDimension(AngularDimension):
    """
    Arc is defined by start- and endpoint on arc and the center point, or by three points lying on the arc if acr3points
    is True. Measured length goes from start- to endpoint. The dimension line goes through the dimlinepos.

    """

    def __init__(
        self,
        pos: UVec,
        center: UVec,
        start: UVec,
        end: UVec,
        arc3points: bool = False,
        dimstyle: str = "Default",
        layer: Optional[str] = None,
        roundval: Optional[int] = None,
    ):
        """
        Args:
            pos: location as (x, y) tuple of dimension line, line goes through this point
            center: center point of arc
            start: start point of arc
            end: end point of arc
            arc3points: if **True** arc is defined by three points on the arc (center, start, end)
            dimstyle: dimstyle name, 'Default' - style is the default value
            layer: dimension line layer, override the default value of dimstyle
            roundval: count of decimal places

        """
        super().__init__(pos, center, start, end, dimstyle, layer, roundval)
        self.arc3points = arc3points

    def _setup(self) -> None:
        super()._setup()
        if self.arc3points:
            self.center = center_of_3points_arc(
                self.center, self.start, self.end
            )

    def _get_extline_start(self, vector: Vec3) -> Vec3:
        return self.center + (
            vector * (self.radius + self.prop("dimextlinegap"))
        )

    def _get_extline_end(self, vector: Vec3) -> Vec3:
        return self.center + (vector * self.pos_radius)

    def _get_dimtext(self) -> str:
        arc_length = (
            (self.end_angle - self.start_angle)
            * self.radius
            * self.prop("scale")
        )
        return self.format_dimtext(arc_length)


class RadialDimension(_DimensionBase):
    """
    Draw a radius dimension line from `target` in direction of `center` with length drawing units. RadiusDimension has
    a special tick!!
    """

    def __init__(
        self,
        center: UVec,
        target: UVec,
        length: float = 1.0,
        dimstyle: str = "Default",
        layer: Optional[str] = None,
        roundval: Optional[int] = None,
    ):
        """
        Args:
            center: center point of radius
            target: target point of radius
            length: length of radius arrow (drawing length)
            dimstyle: dimstyle name, 'Default' - style is the default value
            layer: dimension line layer, override the default value of dimstyle
            roundval: count of decimal places

        """
        super().__init__(dimstyle, layer, roundval)  # type: ignore
        self.center = Vec3(center)
        self.target = Vec3(target)
        self.length = float(length)

    def _setup(self) -> None:
        self.target_vector = (
            self.target - self.center
        ).normalize()  # type: Vec3
        self.radius = distance(self.center, self.target)  # type: float

    def render(self, layout: "GenericLayoutType") -> None:
        """build dimension line object with basic dxf entities"""
        self._setup()
        self._draw_dimension_line(layout)
        self._draw_dimension_text(layout)
        self._draw_ticks(layout)

    def _draw_dimension_line(self, layout: "GenericLayoutType") -> None:
        start_point = self.center + (
            self.target_vector * (self.radius - self.length)
        )
        layout.add_line(
            start=start_point,
            end=self.target,
            dxfattribs={
                "color": self.prop("dimlinecolor"),
                "layer": self.prop("layer"),
            },
        )

    def _draw_dimension_text(self, layout: "GenericLayoutType") -> None:
        layout.add_text(
            text=self._get_dimtext(),
            dxfattribs={
                "height": self.prop("height"),
                "rotation": self.target_vector.angle_deg,
                "layer": self.prop("layer"),
                "style": self.prop("style"),
                "color": self.prop("textcolor"),
            },
        ).set_placement(
            self._get_insert_point(), align=TextEntityAlignment.MIDDLE_RIGHT
        )

    def _get_insert_point(self) -> Vec3:
        return self.target - (
            self.target_vector * (self.length + self.prop("textabove"))
        )

    def _get_dimtext(self) -> str:
        return self.format_dimtext(self.radius * self.prop("scale"))

    def _draw_ticks(self, layout: "GenericLayoutType") -> None:
        layout.add_blockref(
            insert=self.target,
            name="DIMTICK_RADIUS",
            dxfattribs={
                "rotation": self.target_vector.angle_deg + 180,
                "xscale": self.prop("tickfactor"),
                "yscale": self.prop("tickfactor"),
                "layer": self.prop("layer"),
            },
        )


def center_of_3points_arc(point1: UVec, point2: UVec, point3: UVec) -> Vec2:
    """
    Calc center point of 3 point arc. ConstructionCircle is defined by 3 points
    on the circle: point1, point2 and point3.
    """
    ray1 = ConstructionRay(point1, point2)
    ray2 = ConstructionRay(point1, point3)
    midpoint1 = lerp(point1, point2)
    midpoint2 = lerp(point1, point3)
    center_ray1 = ray1.orthogonal(midpoint1)
    center_ray2 = ray2.orthogonal(midpoint2)
    return center_ray1.intersect(center_ray2)



#  Copyright (c) 2022-2024, Manfred Moitzi
#  License: MIT License
import enum
from ezdxf.version import __version__

# SAT Export Requirements for Autodesk Products
# ---------------------------------------------
# Script to create test files:
# examples/acistools/create_3dsolid_cube.py

# DXF R2000, R2004, R2007, R2010: OK, tested with TrueView 2022
# ACIS version 700
# ACIS version string: "ACIS 32.0 NT"
# record count: 0, not required
# body count: 1, required
# ASM header: no
# end-marker: "End-of-ACIS-data"

# DXF R2004, R2007, R2010: OK, tested with TrueView 2022
# ACIS version 20800
# ACIS version string: "ACIS 208.00 NT"
# record count: 0, not required
# body count: n + 1 (asm-header), required
# ASM header: "208.0.4.7009"
# end-marker: "End-of-ACIS-data"

# SAB Export Requirements for Autodesk Products
# ---------------------------------------------
# DXF R2013, R2018: OK, tested with TrueView 2022
# ACIS version 21800
# ACIS version string: "ACIS 208.00 NT"
# record count: 0, not required
# body count: n + 1  (asm-header), required
# ASM header: "208.0.4.7009"
# end-marker: "End-of-ASM-data"

ACIS_VERSION = {
    400: "ACIS 4.00 NT",  # DXF R2000, no asm header - only R2000
    700: "ACIS 32.0 NT",  # DXF R2000-R2010, no asm header
    20800: "ACIS 208.00 NT",  # DXF R2013 with asm-header, asm-end-marker
    21800: "ACIS 218.00 NT",  # DXF R2013 with asm-header, asm-end-marker
    22300: "ACIS 223.00 NT",  # DXF R2018 with asm-header, asm-end-marker
}
ASM_VERSION = {
    20800: "208.0.4.7009",  # DXF R2004, R2007, R2010
    21800: "208.0.4.7009",  # DXF R2013, default version for R2013 and R2018
    22300: "222.0.0.1700",  # DXF R2018
}
EZDXF_BUILDER_ID = f"ezdxf v{__version__} ACIS Builder"
MIN_EXPORT_VERSION = 700

# ACIS version 700 is the default version for DXF R2000, R2004, R2007 and R2010 (SAT)
# ACIS version 21800 is the default version for DXF R2013 and R2018 (SAB)
DEFAULT_SAT_VERSION = 700
DEFAULT_SAB_VERSION = 21800

DATE_FMT = "%a %b %d %H:%M:%S %Y"
END_OF_ACIS_DATA_SAT = "End-of-ACIS-data"
END_OF_ACIS_DATA_SAB = b"\x0e\x03End\x0e\x02of\x0e\x04ACIS\x0d\x04data"
END_OF_ASM_DATA_SAT = "End-of-ASM-data"
END_OF_ASM_DATA_SAB = b"\x0e\x03End\x0e\x02of\x0e\x03ASM\x0d\x04data"
BEGIN_OF_ACIS_HISTORY_DATA = "Begin-of-ACIS-History-data"
END_OF_ACIS_HISTORY_DATA = "End-of-ACIS-History-data"
DATA_END_MARKERS = (
    END_OF_ACIS_DATA_SAT,
    BEGIN_OF_ACIS_HISTORY_DATA,
    END_OF_ASM_DATA_SAT,
)
NULL_PTR_NAME = "null-ptr"
NONE_ENTITY_NAME = "none-entity"
NOR_TOL = 1e-10
RES_TOL = 9.9999999999999995e-7

BOOL_SPECIFIER = {
    "forward": True,
    "forward_v": True,
    "reversed": False,
    "reversed_v": False,
    "single": True,
    "double": False,
}

ACIS_SIGNATURE = b"ACIS BinaryFile"  # DXF R2013/R2018
ASM_SIGNATURE = b"ASM BinaryFile4"  # DXF R2018
SIGNATURES = [ACIS_SIGNATURE, ASM_SIGNATURE]


def is_valid_export_version(version: int):
    return version >= MIN_EXPORT_VERSION and version in ACIS_VERSION


class Tags(enum.IntEnum):
    NO_TYPE = 0x00
    BYTE = 0x01  # not used in files!
    CHAR = 0x02  # not used in files!
    SHORT = 0x03  # not used in files!
    INT = 0x04  # 32-bit signed integer
    FLOAT = 0x05  # not used in files!
    DOUBLE = 0x06  # 64-bit double precision floating point value
    STR = 0x07  # count is the following 8-bit uchar
    STR2 = 0x08  # not used in files!
    STR3 = 0x09  # not used in files!

    # bool value for reversed, double, I - depends on context
    BOOL_TRUE = 0x0A

    # bool value forward, single, forward_v - depends on context
    BOOL_FALSE = 0x0B
    POINTER = 0x0C
    ENTITY_TYPE = 0x0D
    ENTITY_TYPE_EX = 0x0E
    SUBTYPE_START = 0x0F
    SUBTYPE_END = 0x10
    RECORD_END = 0x11
    LITERAL_STR = 0x12  # count ia a 32-bit uint, see transform entity
    LOCATION_VEC = 0x13  # vector (3 doubles)
    DIRECTION_VEC = 0x14  # vector (3 doubles)

    # Enumeration are stored as strings in SAT and ints in SAB.
    # It's not possible to translate SAT enums (strings) to SAB enums (int) and
    # vice versa without knowing the implementation details. Each enumeration
    # is specific to the class where it is used.
    ENUM = 0x15
    # 0x16: ???
    UNKNOWN_0x17 = 0x17  # double


# entity type structure:
# 0x0D 0x04 (char count of) "body" = SAT "body"
# 0x0E 0x05 "plane" 0x0D 0x07 "surface" = SAT "plane-surface"
# 0x0E 0x06 "ref_vt" 0x0E 0x03 "eye" 0x0D 0x06 "attrib" = SAT "ref_vt-eye-attrib"


class Flags(enum.IntFlag):
    HAS_HISTORY = 1


class AcisException(Exception):
    pass


class InvalidLinkStructure(AcisException):
    pass


class ParsingError(AcisException):
    pass


class ExportError(AcisException):
    pass


class EndOfAcisData(AcisException):
    pass


class Features:
    LAW_SPL = 400
    CONE_SCALING = 400
    LOFT_LAW = 400
    REF_MIN_UV_GRID = 400
    VBLEND_AUTO = 400
    BL_ENV_SF = 400
    ELLIPSE_OFFSET = 500
    TOL_MODELING = 500
    APPROX_SUMMARY = 500
    TAPER_SCALING = 500
    LAZY_B_SPLINE = 500
    DM_MULTI_SURF = 500
    GA_COPY_ACTION = 600
    DM_MULTI_SURF_COLOR = 600
    RECAL_SKIN_ERROR = 520
    TAPER_U_RULED = 600
    DM_60 = 600
    LOFT_PCURVE = 600
    EELIST_OWNER = 600
    ANNO_HOOKED = 700
    PATTERN = 700
    ENTITY_TAGS = 700
    AT = 700
    NET_LAW = 700
    STRINGLESS_HISTORY = 700



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from enum import IntEnum, IntFlag, Enum, auto
from ezdxf.lldxf import const


class TextHAlign(IntEnum):
    """Enumeration for DXF attribute: :attr:`ezdxf.entities.Text.dxf.halign`"""

    LEFT = const.LEFT
    CENTER = const.CENTER
    RIGHT = const.RIGHT
    ALIGNED = const.ALIGNED
    MIDDLE = 4
    FIT = const.FIT


class TextVAlign(IntEnum):
    """Enumeration for DXF attribute: :attr:`ezdxf.entities.Text.dxf.valign`"""

    BASELINE = const.BASELINE
    BOTTOM = const.BOTTOM
    MIDDLE = const.MIDDLE
    TOP = const.TOP


# noinspection PyArgumentList
class TextEntityAlignment(Enum):
    """Text alignment enum for the :class:`~ezdxf.entities.Text`,
    :class:`~ezdxf.entities.Attrib` and :class:`~ezdxf.entities.AttDef`
    entities.
    """

    LEFT = auto()
    CENTER = auto()
    RIGHT = auto()
    ALIGNED = auto()
    MIDDLE = auto()
    FIT = auto()
    BOTTOM_LEFT = auto()
    BOTTOM_CENTER = auto()
    BOTTOM_RIGHT = auto()
    MIDDLE_LEFT = auto()
    MIDDLE_CENTER = auto()
    MIDDLE_RIGHT = auto()
    TOP_LEFT = auto()
    TOP_CENTER = auto()
    TOP_RIGHT = auto()


MAP_TEXT_ENUM_TO_ALIGN_FLAGS = {
    TextEntityAlignment.LEFT: (TextHAlign.LEFT, TextVAlign.BASELINE),
    TextEntityAlignment.CENTER: (TextHAlign.CENTER, TextVAlign.BASELINE),
    TextEntityAlignment.RIGHT: (TextHAlign.RIGHT, TextVAlign.BASELINE),
    TextEntityAlignment.ALIGNED: (TextHAlign.ALIGNED, TextVAlign.BASELINE),
    TextEntityAlignment.MIDDLE: (TextHAlign.MIDDLE, TextVAlign.BASELINE),
    TextEntityAlignment.FIT: (TextHAlign.FIT, TextVAlign.BASELINE),
    TextEntityAlignment.BOTTOM_LEFT: (TextHAlign.LEFT, TextVAlign.BOTTOM),
    TextEntityAlignment.BOTTOM_CENTER: (TextHAlign.CENTER, TextVAlign.BOTTOM),
    TextEntityAlignment.BOTTOM_RIGHT: (TextHAlign.RIGHT, TextVAlign.BOTTOM),
    TextEntityAlignment.MIDDLE_LEFT: (TextHAlign.LEFT, TextVAlign.MIDDLE),
    TextEntityAlignment.MIDDLE_CENTER: (TextHAlign.CENTER, TextVAlign.MIDDLE),
    TextEntityAlignment.MIDDLE_RIGHT: (TextHAlign.RIGHT, TextVAlign.MIDDLE),
    TextEntityAlignment.TOP_LEFT: (TextHAlign.LEFT, TextVAlign.TOP),
    TextEntityAlignment.TOP_CENTER: (TextHAlign.CENTER, TextVAlign.TOP),
    TextEntityAlignment.TOP_RIGHT: (TextHAlign.RIGHT, TextVAlign.TOP),
}
MAP_TEXT_ALIGN_FLAGS_TO_ENUM = dict(
    (flags, enum) for enum, flags in MAP_TEXT_ENUM_TO_ALIGN_FLAGS.items()
)

# Used by legacy add-ons MText and Table!
MAP_STRING_ALIGN_TO_FLAGS = {
    "LEFT": (TextHAlign.LEFT, TextVAlign.BASELINE),
    "CENTER": (TextHAlign.CENTER, TextVAlign.BASELINE),
    "RIGHT": (TextHAlign.RIGHT, TextVAlign.BASELINE),
    "ALIGNED": (TextHAlign.ALIGNED, TextVAlign.BASELINE),
    "MIDDLE": (TextHAlign.MIDDLE, TextVAlign.BASELINE),
    "FIT": (TextHAlign.FIT, TextVAlign.BASELINE),
    "BOTTOM_LEFT": (TextHAlign.LEFT, TextVAlign.BOTTOM),
    "BOTTOM_CENTER": (TextHAlign.CENTER, TextVAlign.BOTTOM),
    "BOTTOM_RIGHT": (TextHAlign.RIGHT, TextVAlign.BOTTOM),
    "MIDDLE_LEFT": (TextHAlign.LEFT, TextVAlign.MIDDLE),
    "MIDDLE_CENTER": (TextHAlign.CENTER, TextVAlign.MIDDLE),
    "MIDDLE_RIGHT": (TextHAlign.RIGHT, TextVAlign.MIDDLE),
    "TOP_LEFT": (TextHAlign.LEFT, TextVAlign.TOP),
    "TOP_CENTER": (TextHAlign.CENTER, TextVAlign.TOP),
    "TOP_RIGHT": (TextHAlign.RIGHT, TextVAlign.TOP),
}
MAP_FLAGS_TO_STRING_ALIGN = dict(
    (flags, name) for name, flags in MAP_STRING_ALIGN_TO_FLAGS.items()
)


class MTextEntityAlignment(IntEnum):
    """Text alignment enum for the :class:`~ezdxf.entities.MText` entity."""

    TOP_LEFT = const.MTEXT_TOP_LEFT
    TOP_CENTER = const.MTEXT_TOP_CENTER
    TOP_RIGHT = const.MTEXT_TOP_RIGHT
    MIDDLE_LEFT = const.MTEXT_MIDDLE_LEFT
    MIDDLE_CENTER = const.MTEXT_MIDDLE_CENTER
    MIDDLE_RIGHT = const.MTEXT_MIDDLE_RIGHT
    BOTTOM_LEFT = const.MTEXT_BOTTOM_LEFT
    BOTTOM_CENTER = const.MTEXT_BOTTOM_CENTER
    BOTTOM_RIGHT = const.MTEXT_BOTTOM_RIGHT


MAP_MTEXT_ALIGN_TO_FLAGS = {
    MTextEntityAlignment.TOP_LEFT: (TextHAlign.LEFT, TextVAlign.TOP),
    MTextEntityAlignment.TOP_CENTER: (TextHAlign.CENTER, TextVAlign.TOP),
    MTextEntityAlignment.TOP_RIGHT: (TextHAlign.RIGHT, TextVAlign.TOP),
    MTextEntityAlignment.MIDDLE_LEFT: (TextHAlign.LEFT, TextVAlign.MIDDLE),
    MTextEntityAlignment.MIDDLE_CENTER: (TextHAlign.CENTER, TextVAlign.MIDDLE),
    MTextEntityAlignment.MIDDLE_RIGHT: (TextHAlign.RIGHT, TextVAlign.MIDDLE),
    MTextEntityAlignment.BOTTOM_LEFT: (TextHAlign.LEFT, TextVAlign.BOTTOM),
    MTextEntityAlignment.BOTTOM_CENTER: (TextHAlign.CENTER, TextVAlign.BOTTOM),
    MTextEntityAlignment.BOTTOM_RIGHT: (TextHAlign.RIGHT, TextVAlign.BOTTOM),
}


class MTextParagraphAlignment(IntEnum):
    DEFAULT = 0
    LEFT = 1
    RIGHT = 2
    CENTER = 3
    JUSTIFIED = 4
    DISTRIBUTED = 5


class MTextFlowDirection(IntEnum):
    LEFT_TO_RIGHT = const.MTEXT_LEFT_TO_RIGHT
    TOP_TO_BOTTOM = const.MTEXT_TOP_TO_BOTTOM
    BY_STYLE = const.MTEXT_BY_STYLE


class MTextLineAlignment(IntEnum):  # exclusive state
    BOTTOM = 0
    MIDDLE = 1
    TOP = 2


class MTextStroke(IntFlag):
    """Combination of flags is supported: UNDERLINE + STRIKE_THROUGH"""

    UNDERLINE = 1
    STRIKE_THROUGH = 2
    OVERLINE = 4


class MTextLineSpacing(IntEnum):
    AT_LEAST = const.MTEXT_AT_LEAST
    EXACT = const.MTEXT_EXACT


class MTextBackgroundColor(IntEnum):
    OFF = const.MTEXT_BG_OFF
    COLOR = const.MTEXT_BG_COLOR
    WINDOW = const.MTEXT_BG_WINDOW_COLOR
    CANVAS = const.MTEXT_BG_CANVAS_COLOR


class InsertUnits(IntEnum):
    Unitless = 0
    Inches = 1
    Feet = 2
    Miles = 3
    Millimeters = 4
    Centimeters = 5
    Meters = 6
    Kilometers = 7
    Microinches = 8
    Mils = 9
    Yards = 10
    Angstroms = 11
    Nanometers = 12
    Microns = 13
    Decimeters = 14
    Decameters = 15
    Hectometers = 16
    Gigameters = 17
    AstronomicalUnits = 18
    Lightyears = 19
    Parsecs = 20
    USSurveyFeet = 21
    USSurveyInch = 22
    USSurveyYard = 23
    USSurveyMile = 24


class Measurement(IntEnum):
    Imperial = 0
    Metric = 1


class LengthUnits(IntEnum):
    Scientific = 1
    Decimal = 2
    Engineering = 3
    Architectural = 4
    Fractional = 5


class AngularUnits(IntEnum):
    DecimalDegrees = 0
    DegreesMinutesSeconds = 1
    Grad = 2
    Radians = 3


class SortEntities(IntFlag):
    DISABLE = 0
    SELECTION = 1  # 1 = Sorts for object selection
    SNAP = 2  # 2 = Sorts for object snap
    REDRAW = 4  # 4 = Sorts for redraws; obsolete
    MSLIDE = 8  # 8 = Sorts for MSLIDE command slide creation; obsolete
    REGEN = 16  # 16 = Sorts for REGEN commands
    PLOT = 32  # 32 = Sorts for plotting
    POSTSCRIPT = 64  # 64 = Sorts for PostScript output; obsolete


class ACI(IntEnum):
    """AutoCAD Color Index"""

    BYBLOCK = 0
    BYLAYER = 256
    BYOBJECT = 257
    RED = 1
    YELLOW = 2
    GREEN = 3
    CYAN = 4
    BLUE = 5
    MAGENTA = 6
    BLACK = 7
    WHITE = 7
    GRAY = 8
    LIGHT_GRAY = 9


class EndCaps(IntEnum):
    """ Lineweight end caps setting for new objects. """
    NONE = 0
    ROUND = 1
    ANGLE = 2
    SQUARE = 3


class JoinStyle(IntEnum):
    """ Lineweight joint setting for new objects. """
    NONE = 0
    ROUND = 1
    ANGLE = 2
    FLAT = 3



# Copyright (c) 2018-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, cast, Optional
import math
from ezdxf.math import Vec3, Vec2, UVec, ConstructionRay, UCS
from ezdxf.render.arrows import ARROWS, connection_point
from ezdxf.entities.dimstyleoverride import DimStyleOverride

from .dim_base import (
    BaseDimensionRenderer,
    LengthMeasurement,
    Measurement,
    compile_mtext,
    order_leader_points,
)

if TYPE_CHECKING:
    from ezdxf.entities import Dimension
    from ezdxf.eztypes import GenericLayoutType


class LinearDimension(BaseDimensionRenderer):
    """Linear dimension line renderer, used for horizontal, vertical, rotated
    and aligned DIMENSION entities.

    Args:
        dimension: DXF entity DIMENSION
        ucs: user defined coordinate system
        override: dimension style override management object

    """

    def __init__(
        self,
        dimension: Dimension,
        ucs: Optional[UCS] = None,
        override: Optional[DimStyleOverride] = None,
    ):
        super().__init__(dimension, ucs, override)
        measurement = self.measurement
        if measurement.text_movement_rule == 0:
            # moves the dimension line with dimension text, this makes no sense
            # for ezdxf (just set `base` argument)
            measurement.text_movement_rule = 2

        self.oblique_angle: float = self.dimension.get_dxf_attrib(
            "oblique_angle", 90
        )
        self.dim_line_angle: float = self.dimension.get_dxf_attrib("angle", 0)
        self.dim_line_angle_rad: float = math.radians(self.dim_line_angle)
        self.ext_line_angle: float = self.dim_line_angle + self.oblique_angle
        self.ext_line_angle_rad: float = math.radians(self.ext_line_angle)

        # text is aligned to dimension line
        measurement.text_rotation = self.dim_line_angle
        # text above extension line, is always aligned with extension lines
        if measurement.text_halign in (3, 4):
            measurement.text_rotation = self.ext_line_angle

        self.ext1_line_start = Vec2(self.dimension.dxf.defpoint2)
        self.ext2_line_start = Vec2(self.dimension.dxf.defpoint3)

        ext1_ray = ConstructionRay(
            self.ext1_line_start, angle=self.ext_line_angle_rad
        )
        ext2_ray = ConstructionRay(
            self.ext2_line_start, angle=self.ext_line_angle_rad
        )
        dim_line_ray = ConstructionRay(
            self.dimension.dxf.defpoint, angle=self.dim_line_angle_rad
        )

        self.dim_line_start: Vec2 = dim_line_ray.intersect(ext1_ray)
        self.dim_line_end: Vec2 = dim_line_ray.intersect(ext2_ray)
        self.dim_line_center: Vec2 = self.dim_line_start.lerp(self.dim_line_end)

        if self.dim_line_start == self.dim_line_end:
            self.dim_line_vec = Vec2.from_angle(self.dim_line_angle_rad)
        else:
            self.dim_line_vec = (
                self.dim_line_end - self.dim_line_start
            ).normalize()

        # set dimension defpoint to expected location - 3D vertex required!
        self.dimension.dxf.defpoint = Vec3(self.dim_line_start)

        raw_measurement = (self.dim_line_end - self.dim_line_start).magnitude
        measurement.update(raw_measurement)

        # only for linear dimension in multi point mode
        self.multi_point_mode = self.dim_style.pop("multi_point_mode", False)

        # 1 .. move wide text up
        # 2 .. move wide text down
        # None .. ignore
        self.move_wide_text: Optional[bool] = self.dim_style.pop(
            "move_wide_text", None
        )

        # actual text width in drawing units
        self._total_text_width: float = 0

        # arrows
        self.required_arrows_space: float = (
            2 * self.arrows.arrow_size + measurement.text_gap
        )
        self.arrows_outside: bool = self.required_arrows_space > raw_measurement

        # text location and rotation
        if measurement.text:
            # text width and required space
            self._total_text_width = self.total_text_width()
            if self.tol.has_limits:
                # limits show the upper and lower limit of the measurement as
                # stacked values and with the size of tolerances
                self.tol.update_limits(self.measurement.value)

            if self.multi_point_mode:
                # ezdxf has total control about vertical text position in multi
                # point mode
                measurement.text_vertical_position = 0.0

            if (
                measurement.text_valign == 0
                and abs(measurement.text_vertical_position) < 0.7
            ):
                # vertical centered text needs also space for arrows
                required_space = (
                    self._total_text_width + 2 * self.arrows.arrow_size
                )
            else:
                required_space = self._total_text_width
            measurement.is_wide_text = required_space > raw_measurement

            if not measurement.force_text_inside:
                # place text outside if wide text and not forced inside
                measurement.text_is_outside = measurement.is_wide_text
            elif measurement.is_wide_text and measurement.text_halign < 3:
                # center wide text horizontal
                measurement.text_halign = 0

            # use relative text shift to move wide text up or down in multi
            # point mode
            if (
                self.multi_point_mode
                and measurement.is_wide_text
                and self.move_wide_text
            ):
                shift_value = measurement.text_height + measurement.text_gap
                if self.move_wide_text == 1:  # move text up
                    measurement.text_shift_v = shift_value
                    if (
                        measurement.vertical_placement == -1
                    ):  # text below dimension line
                        # shift again
                        measurement.text_shift_v += shift_value
                elif self.move_wide_text == 2:  # move text down
                    measurement.text_shift_v = -shift_value
                    if (
                        measurement.vertical_placement == 1
                    ):  # text above dimension line
                        # shift again
                        measurement.text_shift_v -= shift_value

            # get final text location - no altering after this line
            measurement.text_location = self.get_text_location()

            # text rotation override
            rotation: float = measurement.text_rotation
            if measurement.user_text_rotation is not None:
                rotation = measurement.user_text_rotation
            elif (
                measurement.text_is_outside
                and measurement.text_outside_horizontal
            ):
                rotation = 0.0
            elif (
                measurement.text_is_inside
                and measurement.text_inside_horizontal
            ):
                rotation = 0.0
            measurement.text_rotation = rotation

            text_box = self.init_text_box()
            self.geometry.set_text_box(text_box)
            if measurement.has_leader:
                p1, p2, *_ = text_box.corners
                self.leader1, self.leader2 = order_leader_points(
                    self.dim_line_center, p1, p2
                )
                # not exact what BricsCAD (AutoCAD) expect, but close enough
                self.dimension.dxf.text_midpoint = self.leader1
            else:
                # write final text location into DIMENSION entity
                self.dimension.dxf.text_midpoint = measurement.text_location

    def init_measurement(self, color: int, scale: float) -> Measurement:
        return LengthMeasurement(
            self.dim_style, self.default_color, self.dim_scale
        )

    def render(self, block: GenericLayoutType) -> None:
        """Main method to create dimension geometry of basic DXF entities in the
        associated BLOCK layout.

        Args:
            block: target BLOCK for rendering

        """
        # call required to setup some requirements
        super().render(block)

        # add extension line 1
        ext_lines = self.extension_lines
        measurement = self.measurement
        if not ext_lines.suppress1:
            above_ext_line1 = measurement.text_halign == 3
            start, end = self.extension_line_points(
                self.ext1_line_start, self.dim_line_start, above_ext_line1
            )
            self.add_line(start, end, dxfattribs=ext_lines.dxfattribs(1))

        # add extension line 2
        if not ext_lines.suppress2:
            above_ext_line2 = measurement.text_halign == 4
            start, end = self.extension_line_points(
                self.ext2_line_start, self.dim_line_end, above_ext_line2
            )
            self.add_line(start, end, dxfattribs=ext_lines.dxfattribs(2))

        # add arrow symbols (block references), also adjust dimension line start
        # and end point
        dim_line_start, dim_line_end = self.add_arrows()

        # add dimension line
        self.add_dimension_line(dim_line_start, dim_line_end)

        # add measurement text as last entity to see text fill properly
        if measurement.text:
            if self.geometry.supports_dxf_r2000:
                text = compile_mtext(measurement, self.tol)
            else:
                text = measurement.text
            self.add_measurement_text(
                text, measurement.text_location, measurement.text_rotation
            )
            if measurement.has_leader:
                self.add_leader(
                    self.dim_line_center, self.leader1, self.leader2
                )

        # add POINT entities at definition points
        self.geometry.add_defpoints(
            [self.dim_line_start, self.ext1_line_start, self.ext2_line_start]
        )

    def get_text_location(self) -> Vec2:
        """Get text midpoint in UCS from user defined location or default text
        location.

        """
        # apply relative text shift as user location override without leader
        measurement = self.measurement
        if measurement.has_relative_text_movement:
            location = self.default_text_location()
            location = measurement.apply_text_shift(
                location, measurement.text_rotation
            )
            self.location_override(location)

        if measurement.user_location is not None:
            location = measurement.user_location
            if measurement.relative_user_location:
                location = self.dim_line_center + location
            # define overridden text location as outside
            measurement.text_is_outside = True
        else:
            location = self.default_text_location()

        return location

    def default_text_location(self) -> Vec2:
        """Calculate default text location in UCS based on `self.text_halign`,
        `self.text_valign` and `self.text_outside`

        """
        start = self.dim_line_start
        end = self.dim_line_end
        measurement = self.measurement
        halign = measurement.text_halign
        # positions the text above and aligned with the first/second extension line
        ext_lines = self.extension_lines
        if halign in (3, 4):
            # horizontal location
            hdist = measurement.text_gap + measurement.text_height / 2.0
            hvec = self.dim_line_vec * hdist
            location = (start if halign == 3 else end) - hvec
            # vertical location
            vdist = ext_lines.extension_above + self._total_text_width / 2.0
            location += Vec2.from_deg_angle(self.ext_line_angle).normalize(
                vdist
            )
        else:
            # relocate outside text to center location
            if measurement.text_is_outside:
                halign = 0

            if halign == 0:
                location = self.dim_line_center  # center of dimension line
            else:
                hdist = (
                    self._total_text_width / 2.0
                    + self.arrows.arrow_size
                    + measurement.text_gap
                )
                if (
                    halign == 1
                ):  # positions the text next to the first extension line
                    location = start + (self.dim_line_vec * hdist)
                else:  # positions the text next to the second extension line
                    location = end - (self.dim_line_vec * hdist)

            if measurement.text_is_outside:  # move text up
                vdist = (
                    ext_lines.extension_above
                    + measurement.text_gap
                    + measurement.text_height / 2.0
                )
            else:
                # distance from extension line to text midpoint
                vdist = measurement.text_vertical_distance()
            location += self.dim_line_vec.orthogonal().normalize(vdist)

        return location

    def add_arrows(self) -> tuple[Vec2, Vec2]:
        """
        Add arrows or ticks to dimension.

        Returns: dimension line connection points

        """
        arrows = self.arrows
        attribs = arrows.dxfattribs()
        start = self.dim_line_start
        end = self.dim_line_end
        outside = self.arrows_outside
        arrow1 = not arrows.suppress1
        arrow2 = not arrows.suppress2

        if arrows.tick_size > 0.0:  # oblique stroke, but double the size
            if arrow1:
                self.add_blockref(
                    ARROWS.oblique,
                    insert=start,
                    rotation=self.dim_line_angle,
                    scale=arrows.tick_size * 2.0,
                    dxfattribs=attribs,
                )
            if arrow2:
                self.add_blockref(
                    ARROWS.oblique,
                    insert=end,
                    rotation=self.dim_line_angle,
                    scale=arrows.tick_size * 2.0,
                    dxfattribs=attribs,
                )
        else:
            scale = arrows.arrow_size
            start_angle = self.dim_line_angle + 180.0
            end_angle = self.dim_line_angle
            if outside:
                start_angle, end_angle = end_angle, start_angle

            if arrow1:
                self.add_blockref(
                    arrows.arrow1_name,
                    insert=start,
                    scale=scale,
                    rotation=start_angle,
                    dxfattribs=attribs,
                )  # reverse
            if arrow2:
                self.add_blockref(
                    arrows.arrow2_name,
                    insert=end,
                    scale=scale,
                    rotation=end_angle,
                    dxfattribs=attribs,
                )

            if not outside:
                # arrows inside extension lines: adjust connection points for
                # the remaining dimension line
                if arrow1:
                    start = connection_point(
                        arrows.arrow1_name, start, scale, start_angle
                    )
                if arrow2:
                    end = connection_point(
                        arrows.arrow2_name, end, scale, end_angle
                    )
            else:
                # add additional extension lines to arrows placed outside of
                # dimension extension lines
                self.add_arrow_extension_lines()
        return start, end

    def add_arrow_extension_lines(self):
        """Add extension lines to arrows placed outside of dimension extension
        lines. Called by `self.add_arrows()`.

        """

        def has_arrow_extension(name: str) -> bool:
            return (
                (name is not None)
                and (name in ARROWS)
                and (name not in ARROWS.ORIGIN_ZERO)
            )

        attribs = self.dimension_line.dxfattribs()
        arrows = self.arrows
        start = self.dim_line_start
        end = self.dim_line_end
        arrow_size = arrows.arrow_size

        if not arrows.suppress1 and has_arrow_extension(arrows.arrow1_name):
            self.add_line(
                start - self.dim_line_vec * arrow_size,
                start - self.dim_line_vec * (2 * arrow_size),
                dxfattribs=attribs,
            )

        if not arrows.suppress2 and has_arrow_extension(arrows.arrow2_name):
            self.add_line(
                end + self.dim_line_vec * arrow_size,
                end + self.dim_line_vec * (2 * arrow_size),
                dxfattribs=attribs,
            )

    def add_measurement_text(
        self, dim_text: str, pos: Vec2, rotation: float
    ) -> None:
        """Add measurement text to dimension BLOCK.

        Args:
            dim_text: dimension text
            pos: text location
            rotation: text rotation in degrees

        """
        self.add_text(dim_text, pos, rotation, dict())

    def add_dimension_line(self, start: Vec2, end: Vec2) -> None:
        """Add dimension line to dimension BLOCK, adds extension DIMDLE if
        required, and uses DIMSD1 or DIMSD2 to suppress first or second part of
        dimension line. Removes line parts hidden by dimension text.

        Args:
            start: dimension line start
            end: dimension line end

        """
        dim_line = self.dimension_line
        arrows = self.arrows
        extension = self.dim_line_vec * dim_line.extension
        ticks = arrows.has_ticks
        if ticks or ARROWS.has_extension_line(arrows.arrow1_name):
            start = start - extension
        if ticks or ARROWS.has_extension_line(arrows.arrow2_name):
            end = end + extension

        attribs = dim_line.dxfattribs()

        if dim_line.suppress1 or dim_line.suppress2:
            # TODO: results not as expected, but good enough
            # center should take into account text location
            center = start.lerp(end)
            if not dim_line.suppress1:
                self.add_line(
                    start, center, dxfattribs=attribs, remove_hidden_lines=True
                )
            if not dim_line.suppress2:
                self.add_line(
                    center, end, dxfattribs=attribs, remove_hidden_lines=True
                )
        else:
            self.add_line(
                start, end, dxfattribs=attribs, remove_hidden_lines=True
            )

    def extension_line_points(
        self, start: Vec2, end: Vec2, text_above_extline=False
    ) -> tuple[Vec2, Vec2]:
        """Adjust start and end point of extension line by dimension variables
        DIMEXE, DIMEXO, DIMEXFIX, DIMEXLEN.

        Args:
            start: start point of extension line (measurement point)
            end: end point at dimension line
            text_above_extline: True if text is above and aligned with extension line

        Returns: adjusted start and end point

        """
        if start == end:
            direction = Vec2.from_deg_angle(self.ext_line_angle)
        else:
            direction = (end - start).normalize()
        if self.extension_lines.has_fixed_length:
            start = end - (direction * self.extension_lines.length_below)
        else:
            start = start + direction * self.extension_lines.offset
        extension = self.extension_lines.extension_above
        if text_above_extline:
            extension += self._total_text_width
        end = end + direction * extension
        return start, end

    def transform_ucs_to_wcs(self) -> None:
        """Transforms dimension definition points into WCS or if required into
        OCS.

        Can not be called in __init__(), because inherited classes may be need
        unmodified values.

        """

        def from_ucs(attr, func):
            point = self.dimension.get_dxf_attrib(attr)
            self.dimension.set_dxf_attrib(attr, func(point))

        ucs = self.geometry.ucs
        from_ucs("defpoint", ucs.to_wcs)
        from_ucs("defpoint2", ucs.to_wcs)
        from_ucs("defpoint3", ucs.to_wcs)
        from_ucs("text_midpoint", ucs.to_ocs)
        self.dimension.dxf.angle = ucs.to_ocs_angle_deg(
            self.dimension.dxf.angle
        )


CAN_SUPPRESS_ARROW1 = {
    ARROWS.dot,
    ARROWS.dot_small,
    ARROWS.dot_blank,
    ARROWS.origin_indicator,
    ARROWS.origin_indicator_2,
    ARROWS.dot_smallblank,
    ARROWS.none,
    ARROWS.oblique,
    ARROWS.box_filled,
    ARROWS.box,
    ARROWS.integral,
    ARROWS.architectural_tick,
}


def sort_projected_points(
    points: Iterable[UVec], angle: float = 0
) -> list[Vec2]:
    direction = Vec2.from_deg_angle(angle)
    projected_vectors = [(direction.project(Vec2(p)), p) for p in points]
    return [p for projection, p in sorted(projected_vectors)]


def multi_point_linear_dimension(
    layout: GenericLayoutType,
    base: UVec,
    points: Iterable[UVec],
    angle: float = 0,
    ucs: Optional[UCS] = None,
    avoid_double_rendering: bool = True,
    dimstyle: str = "EZDXF",
    override: Optional[dict] = None,
    dxfattribs=None,
    discard=False,
) -> None:
    """Creates multiple DIMENSION entities for each point pair in `points`.
    Measurement points will be sorted by appearance on the dimension line
    vector.

    Args:
        layout: target layout (model space, paper space or block)
        base: base point, any point on the dimension line vector will do
        points: iterable of measurement points
        angle: dimension line rotation in degrees (0=horizontal, 90=vertical)
        ucs: user defined coordinate system
        avoid_double_rendering: removes first extension line and arrow of
            following DIMENSION entity
        dimstyle: dimension style name
        override: dictionary of overridden dimension style attributes
        dxfattribs: DXF attributes for DIMENSION entities
        discard: discard rendering result for friendly CAD applications like
            BricsCAD to get a native and likely better rendering result.
            (does not work with AutoCAD)

    """

    def suppress_arrow1(dimstyle_override) -> bool:
        arrow_name1, arrow_name2 = dimstyle_override.get_arrow_names()
        if (arrow_name1 is None) or (arrow_name1 in CAN_SUPPRESS_ARROW1):
            return True
        else:
            return False

    points = sort_projected_points(points, angle)
    base = Vec2(base)
    override = override or {}
    override["dimtix"] = 1  # do not place measurement text outside
    override["dimtvp"] = 0  # do not place measurement text outside
    override["multi_point_mode"] = True
    # 1 .. move wide text up; 2 .. move wide text down; None .. ignore
    # moving text down, looks best combined with text fill bg: DIMTFILL = 1
    move_wide_text = 1
    _suppress_arrow1 = False
    first_run = True

    for p1, p2 in zip(points[:-1], points[1:]):
        _override = dict(override)
        _override["move_wide_text"] = move_wide_text
        if avoid_double_rendering and not first_run:
            _override["dimse1"] = 1
            _override["suppress_arrow1"] = _suppress_arrow1

        style = layout.add_linear_dim(
            Vec3(base),
            Vec3(p1),
            Vec3(p2),
            angle=angle,
            dimstyle=dimstyle,
            override=_override,
            dxfattribs=dxfattribs,
        )
        if first_run:
            _suppress_arrow1 = suppress_arrow1(style)

        renderer = cast(LinearDimension, style.render(ucs, discard=discard))
        if renderer.measurement.is_wide_text:
            # after wide text switch moving direction
            if move_wide_text == 1:
                move_wide_text = 2
            else:
                move_wide_text = 1
        else:  # reset to move text up
            move_wide_text = 1
        first_run = False



# Copyright (c) 2014-2019, Manfred Moitzi
# License: MIT License
from binascii import unhexlify
from ezdxf.tools.binarydata import hex_strings_to_bytes
from ezdxf.tools.binarydata import bytes_to_hexstr


def test_hexstr_to_bytes():
    assert unhexlify("FFFF") == b"\xff\xff"


def test_hexstr_data_to_bytes_1():
    assert hex_strings_to_bytes(["FFFF"]) == b"\xff\xff"


def test_hexstr_data_to_bytes_2():
    assert hex_strings_to_bytes(["F0F0", "1A1C"]) == b"\xF0\xF0\x1A\x1C"


def test_bytes_to_hexstr():
    assert bytes_to_hexstr(b"\xff\xff") == "FFFF"



#  Copyright (c) 2023-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import (
    Sequence,
    Optional,
    Iterable,
    Tuple,
    Iterator,
    Callable,
)
from typing_extensions import TypeAlias
import abc

import numpy as np
import PIL.Image
import PIL.ImageDraw
import PIL.ImageOps


from ezdxf.colors import RGB
import ezdxf.bbox

from ezdxf.fonts import fonts
from ezdxf.math import Vec2, Matrix44, BoundingBox2d, AnyVec
from ezdxf.path import make_path, Path
from ezdxf.render import linetypes
from ezdxf.entities import DXFGraphic, Viewport
from ezdxf.tools.text import replace_non_printable_characters
from ezdxf.tools.clipping_portal import (
    ClippingPortal,
    ClippingShape,
    find_best_clipping_shape,
)
from ezdxf.layouts import Layout
from .backend import BackendInterface, BkPath2d, BkPoints2d, ImageData
from .config import LinePolicy, TextPolicy, ColorPolicy, Configuration
from .properties import BackendProperties, Filling
from .properties import Properties, RenderContext
from .type_hints import Color
from .unified_text_renderer import UnifiedTextRenderer

PatternKey: TypeAlias = Tuple[str, float]
DrawEntitiesCallback: TypeAlias = Callable[[RenderContext, Iterable[DXFGraphic]], None]

__all__ = ["AbstractPipeline", "RenderPipeline2d"]


class AbstractPipeline(abc.ABC):
    """This drawing pipeline separates the frontend from the backend and implements
    these features:

    - automatically linetype rendering
    - font rendering
    - VIEWPORT rendering
    - foreground color mapping according Frontend.config.color_policy

    The pipeline is organized as concatenated render stages.

    """

    text_engine = UnifiedTextRenderer()
    default_font_face = fonts.FontFace()
    draw_entities: DrawEntitiesCallback

    @abc.abstractmethod
    def set_draw_entities_callback(self, callback: DrawEntitiesCallback) -> None: ...

    @abc.abstractmethod
    def set_config(self, config: Configuration) -> None: ...

    @abc.abstractmethod
    def set_current_entity_handle(self, handle: str) -> None: ...

    @abc.abstractmethod
    def push_clipping_shape(
        self, shape: ClippingShape, transform: Matrix44 | None
    ) -> None: ...

    @abc.abstractmethod
    def pop_clipping_shape(self) -> None: ...

    @abc.abstractmethod
    def draw_viewport(
        self,
        vp: Viewport,
        layout_ctx: RenderContext,
        bbox_cache: Optional[ezdxf.bbox.Cache] = None,
    ) -> None:
        """Draw the content of the given viewport current viewport."""
        ...

    @abc.abstractmethod
    def draw_point(self, pos: AnyVec, properties: Properties) -> None: ...

    @abc.abstractmethod
    def draw_line(self, start: AnyVec, end: AnyVec, properties: Properties): ...

    @abc.abstractmethod
    def draw_solid_lines(
        self, lines: Iterable[tuple[AnyVec, AnyVec]], properties: Properties
    ) -> None: ...

    @abc.abstractmethod
    def draw_path(self, path: Path, properties: Properties): ...

    @abc.abstractmethod
    def draw_filled_paths(
        self,
        paths: Iterable[Path],
        properties: Properties,
    ) -> None: ...

    @abc.abstractmethod
    def draw_filled_polygon(
        self, points: Iterable[AnyVec], properties: Properties
    ) -> None: ...

    @abc.abstractmethod
    def draw_text(
        self,
        text: str,
        transform: Matrix44,
        properties: Properties,
        cap_height: float,
        dxftype: str = "TEXT",
    ) -> None: ...

    @abc.abstractmethod
    def draw_image(self, image_data: ImageData, properties: Properties) -> None: ...

    @abc.abstractmethod
    def finalize(self) -> None: ...

    @abc.abstractmethod
    def set_background(self, color: Color) -> None: ...

    @abc.abstractmethod
    def enter_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        # gets the full DXF properties information
        ...

    @abc.abstractmethod
    def exit_entity(self, entity: DXFGraphic) -> None: ...


class RenderStage2d(abc.ABC):
    next_stage: RenderStage2d

    def set_config(self, config: Configuration) -> None:
        pass

    @abc.abstractmethod
    def draw_point(self, pos: Vec2, properties: Properties) -> None: ...

    @abc.abstractmethod
    def draw_line(self, start: Vec2, end: Vec2, properties: Properties): ...

    @abc.abstractmethod
    def draw_solid_lines(
        self, lines: list[tuple[Vec2, Vec2]], properties: Properties
    ) -> None: ...

    @abc.abstractmethod
    def draw_path(self, path: BkPath2d, properties: Properties): ...

    @abc.abstractmethod
    def draw_filled_paths(
        self,
        paths: list[BkPath2d],
        properties: Properties,
    ) -> None: ...

    @abc.abstractmethod
    def draw_filled_polygon(
        self, points: BkPoints2d, properties: Properties
    ) -> None: ...

    @abc.abstractmethod
    def draw_image(self, image_data: ImageData, properties: Properties) -> None: ...


class RenderPipeline2d(AbstractPipeline):
    """Render pipeline for 2D backends."""

    def __init__(self, backend: BackendInterface):
        self.backend = backend
        self.config = Configuration()
        try:  # request default font face
            self.default_font_face = fonts.font_manager.get_font_face("")
        except fonts.FontNotFoundError:  # no default font found
            # last resort MonospaceFont which renders only "tofu"
            pass
        self.clipping_portal = ClippingPortal()
        self.current_vp_scale = 1.0
        self._current_entity_handle: str = ""
        self._color_mapping: dict[str, str] = dict()
        self._pipeline = self.build_render_pipeline()

    def build_render_pipeline(self) -> RenderStage2d:
        backend_stage = BackendStage2d(
            self.backend, converter=self.get_backend_properties
        )
        linetype_stage = LinetypeStage2d(
            self.config,
            get_ltype_scale=self.get_vp_ltype_scale,
            next_stage=backend_stage,
        )
        clipping_stage = ClippingStage2d(
            self.config, self.clipping_portal, next_stage=linetype_stage
        )
        return clipping_stage

    def get_vp_ltype_scale(self) -> float:
        """The linetype pattern should look the same in all viewports
        regardless of the viewport scale.
        """
        return 1.0 / max(self.current_vp_scale, 0.0001)  # max out at 1:10000

    def get_backend_properties(self, properties: Properties) -> BackendProperties:
        try:
            color = self._color_mapping[properties.color]
        except KeyError:
            color = apply_color_policy(
                properties.color, self.config.color_policy, self.config.custom_fg_color
            )
            self._color_mapping[properties.color] = color
        return BackendProperties(
            color,
            properties.lineweight,
            properties.layer,
            properties.pen,
            self._current_entity_handle,
        )

    def set_draw_entities_callback(self, callback: DrawEntitiesCallback) -> None:
        self.draw_entities = callback

    def set_config(self, config: Configuration) -> None:
        self.backend.configure(config)
        self.config = config
        stage = self._pipeline
        while True:
            stage.set_config(config)
            if not hasattr(stage, "next_stage"):  # BackendStage2d
                return
            stage = stage.next_stage

    def set_current_entity_handle(self, handle: str) -> None:
        assert handle is not None
        self._current_entity_handle = handle

    def push_clipping_shape(
        self, shape: ClippingShape, transform: Matrix44 | None
    ) -> None:
        self.clipping_portal.push(shape, transform)

    def pop_clipping_shape(self) -> None:
        self.clipping_portal.pop()

    def draw_viewport(
        self,
        vp: Viewport,
        layout_ctx: RenderContext,
        bbox_cache: Optional[ezdxf.bbox.Cache] = None,
    ) -> None:
        """Draw the content of the given viewport current viewport."""
        if vp.doc is None:
            return
        try:
            msp_limits = vp.get_modelspace_limits()
        except ValueError:  # modelspace limits not detectable
            return
        if self.enter_viewport(vp):
            self.draw_entities(
                layout_ctx.from_viewport(vp),
                filter_vp_entities(vp.doc.modelspace(), msp_limits, bbox_cache),
            )
            self.exit_viewport()

    def enter_viewport(self, vp: Viewport) -> bool:
        """Set current viewport, returns ``True`` for valid viewports."""
        self.current_vp_scale = vp.get_scale()
        m = vp.get_transformation_matrix()
        clipping_path = make_path(vp)
        if len(clipping_path):
            vertices = clipping_path.control_vertices()
            if clipping_path.has_curves:
                layout = vp.get_layout()
                if isinstance(layout, Layout):
                    # plot paper units:
                    #   0: inches, max sagitta = 1/254 = 0.1 mm
                    #   1: millimeters, max sagitta = 0.1 mm
                    #   2: pixels, max sagitta = 0.1 pixel
                    units = layout.dxf.get("plot_paper_units", 1)
                    max_sagitta = 1.0 / 254.0 if units == 0 else 0.1
                    vertices = list(clipping_path.flattening(max_sagitta))
            clipping_shape = find_best_clipping_shape(vertices)
            self.clipping_portal.push(clipping_shape, m)
            return True
        return False

    def exit_viewport(self):
        self.clipping_portal.pop()
        # Reset viewport scaling: viewports cannot be nested!
        self.current_vp_scale = 1.0

    def draw_text(
        self,
        text: str,
        transform: Matrix44,
        properties: Properties,
        cap_height: float,
        dxftype: str = "TEXT",
    ) -> None:
        """Render text as filled paths."""
        text_policy = self.config.text_policy
        pipeline = self._pipeline

        if not text.strip() or text_policy == TextPolicy.IGNORE:
            return  # no point rendering empty strings
        text = prepare_string_for_rendering(text, dxftype)
        font_face = properties.font
        if font_face is None:
            font_face = self.default_font_face

        try:
            glyph_paths = self.text_engine.get_text_glyph_paths(
                text, font_face, cap_height
            )
        except (RuntimeError, ValueError):
            return
        for p in glyph_paths:
            p.transform_inplace(transform)
        transformed_paths: list[BkPath2d] = glyph_paths

        points: list[Vec2]
        if text_policy == TextPolicy.REPLACE_RECT:
            points = []
            for p in transformed_paths:
                points.extend(p.extents())
            if len(points) < 2:
                return
            rect = BkPath2d.from_vertices(BoundingBox2d(points).rect_vertices())
            pipeline.draw_path(rect, properties)
            return
        if text_policy == TextPolicy.REPLACE_FILL:
            points = []
            for p in transformed_paths:
                points.extend(p.extents())
            if len(points) < 2:
                return
            polygon = BkPoints2d(BoundingBox2d(points).rect_vertices())
            if properties.filling is None:
                properties.filling = Filling()
            pipeline.draw_filled_polygon(polygon, properties)
            return

        if (
            self.text_engine.is_stroke_font(font_face)
            or text_policy == TextPolicy.OUTLINE
        ):
            for text_path in transformed_paths:
                pipeline.draw_path(text_path, properties)
            return

        if properties.filling is None:
            properties.filling = Filling()
        pipeline.draw_filled_paths(transformed_paths, properties)

    def finalize(self) -> None:
        self.backend.finalize()

    def set_background(self, color: Color) -> None:
        self.backend.set_background(color)

    def enter_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        self.backend.enter_entity(entity, properties)

    def exit_entity(self, entity: DXFGraphic) -> None:
        self.backend.exit_entity(entity)

    # Enter render pipeline:
    def draw_point(self, pos: AnyVec, properties: Properties) -> None:
        self._pipeline.draw_point(Vec2(pos), properties)

    def draw_line(self, start: AnyVec, end: AnyVec, properties: Properties):
        self._pipeline.draw_line(Vec2(start), Vec2(end), properties)

    def draw_solid_lines(
        self, lines: Iterable[tuple[AnyVec, AnyVec]], properties: Properties
    ) -> None:
        self._pipeline.draw_solid_lines(
            [(Vec2(s), Vec2(e)) for s, e in lines], properties
        )

    def draw_path(self, path: Path, properties: Properties):
        self._pipeline.draw_path(BkPath2d(path), properties)

    def draw_filled_paths(
        self,
        paths: Iterable[Path],
        properties: Properties,
    ) -> None:
        self._pipeline.draw_filled_paths(list(map(BkPath2d, paths)), properties)

    def draw_filled_polygon(
        self, points: Iterable[AnyVec], properties: Properties
    ) -> None:
        self._pipeline.draw_filled_polygon(BkPoints2d(points), properties)

    def draw_image(self, image_data: ImageData, properties: Properties) -> None:
        self._pipeline.draw_image(image_data, properties)


class ClippingStage2d(RenderStage2d):
    def __init__(
        self,
        config: Configuration,
        clipping_portal: ClippingPortal,
        next_stage: RenderStage2d,
    ):
        self.clipping_portal = clipping_portal
        self.config = config
        self.next_stage = next_stage

    def set_config(self, config: Configuration) -> None:
        self.config = config

    def draw_point(self, pos: Vec2, properties: Properties) -> None:
        if self.clipping_portal.is_active:
            pos = self.clipping_portal.clip_point(pos)
            if pos is None:
                return
        self.next_stage.draw_point(pos, properties)

    def draw_line(self, start: Vec2, end: Vec2, properties: Properties):
        next_stage = self.next_stage
        clipping_portal = self.clipping_portal

        if clipping_portal.is_active:
            for segment in clipping_portal.clip_line(start, end):
                next_stage.draw_line(segment[0], segment[1], properties)
            return
        next_stage.draw_line(start, end, properties)

    def draw_solid_lines(
        self, lines: list[tuple[Vec2, Vec2]], properties: Properties
    ) -> None:
        clipping_portal = self.clipping_portal

        if clipping_portal.is_active:
            cropped_lines: list[tuple[Vec2, Vec2]] = []
            for start, end in lines:
                cropped_lines.extend(clipping_portal.clip_line(start, end))
            lines = cropped_lines
        self.next_stage.draw_solid_lines(lines, properties)

    def draw_path(self, path: BkPath2d, properties: Properties):
        clipping_portal = self.clipping_portal
        next_stage = self.next_stage
        max_sagitta = self.config.max_flattening_distance

        if clipping_portal.is_active:
            for clipped_path in clipping_portal.clip_paths([path], max_sagitta):
                next_stage.draw_path(clipped_path, properties)
            return
        next_stage.draw_path(path, properties)

    def draw_filled_paths(
        self,
        paths: list[BkPath2d],
        properties: Properties,
    ) -> None:
        clipping_portal = self.clipping_portal
        max_sagitta = self.config.max_flattening_distance

        if clipping_portal.is_active:
            paths = clipping_portal.clip_filled_paths(paths, max_sagitta)
        if len(paths) == 0:
            return
        self.next_stage.draw_filled_paths(paths, properties)

    def draw_filled_polygon(self, points: BkPoints2d, properties: Properties) -> None:
        clipping_portal = self.clipping_portal
        next_stage = self.next_stage

        if clipping_portal.is_active:
            for points in clipping_portal.clip_polygon(points):
                if len(points) > 0:
                    next_stage.draw_filled_polygon(points, properties)
            return

        if len(points) > 0:
            next_stage.draw_filled_polygon(points, properties)

    def draw_image(self, image_data: ImageData, properties: Properties) -> None:
        # the outer bounds contain the visible parts of the image for the
        # clip mode "remove inside"
        outer_bounds: list[BkPoints2d] = []
        clipping_portal = self.clipping_portal

        if not clipping_portal.is_active:
            self._draw_image(image_data, outer_bounds, properties)
            return

        # the pixel boundary path can be split into multiple paths
        transform = image_data.flip_matrix() * image_data.transform
        pixel_boundary_path = image_data.pixel_boundary_path
        clipping_paths = _clip_image_polygon(
            clipping_portal, pixel_boundary_path, transform
        )
        if not image_data.remove_outside:
            # remove inside:
            #  detect the visible parts of the image which are not removed by
            #  clipping through viewports or block references
            width, height = image_data.image_size()
            outer_boundary = BkPoints2d(
                Vec2.generate([(0, 0), (width, 0), (width, height), (0, height)])
            )
            outer_bounds = _clip_image_polygon(
                clipping_portal, outer_boundary, transform
            )
        image_data.transform = clipping_portal.transform_matrix(image_data.transform)
        if len(clipping_paths) == 1:
            new_clipping_path = clipping_paths[0]
            if new_clipping_path is not image_data.pixel_boundary_path:
                image_data.pixel_boundary_path = new_clipping_path
                # forced clipping triggered by viewport- or block reference clipping:
                image_data.use_clipping_boundary = True
            self._draw_image(image_data, outer_bounds, properties)
        else:
            for clipping_path in clipping_paths:
                # when clipping path is split into multiple parts:
                #  copy image for each part, not efficient but works
                #  this should be a rare usecase so optimization is not required
                self._draw_image(
                    ImageData(
                        image=image_data.image.copy(),
                        transform=image_data.transform,
                        pixel_boundary_path=clipping_path,
                        use_clipping_boundary=True,
                    ),
                    outer_bounds,
                    properties,
                )

    def _draw_image(
        self,
        image_data: ImageData,
        outer_bounds: list[BkPoints2d],
        properties: Properties,
    ) -> None:
        if image_data.use_clipping_boundary:
            _mask_image(image_data, outer_bounds)
        self.next_stage.draw_image(image_data, properties)


class LinetypeStage2d(RenderStage2d):
    def __init__(
        self,
        config: Configuration,
        get_ltype_scale: Callable[[], float],
        next_stage: RenderStage2d,
    ):
        self.config = config
        self.solid_lines_only = False
        self.next_stage = next_stage
        self.get_ltype_scale = get_ltype_scale
        self.pattern_cache: dict[PatternKey, Sequence[float]] = dict()
        self.set_config(config)

    def set_config(self, config: Configuration) -> None:
        self.config = config
        self.solid_lines_only = config.line_policy == LinePolicy.SOLID

    def pattern(self, properties: Properties) -> Sequence[float]:
        """Returns simplified linetype tuple: on-off sequence"""
        if self.solid_lines_only:
            scale = 0.0
        else:
            scale = properties.linetype_scale * self.get_ltype_scale()

        key: PatternKey = (properties.linetype_name, scale)
        pattern_ = self.pattern_cache.get(key)
        if pattern_ is None:
            pattern_ = self._create_pattern(properties, scale)
            self.pattern_cache[key] = pattern_
        return pattern_

    def _create_pattern(self, properties: Properties, scale: float) -> Sequence[float]:
        if len(properties.linetype_pattern) < 2:
            # Do not return None -> None indicates: "not cached"
            return tuple()

        min_dash_length = self.config.min_dash_length * self.get_ltype_scale()
        pattern = [max(e * scale, min_dash_length) for e in properties.linetype_pattern]
        if len(pattern) % 2:
            pattern.pop()
        return pattern

    def draw_point(self, pos: Vec2, properties: Properties) -> None:
        self.next_stage.draw_point(pos, properties)

    def draw_line(self, start: Vec2, end: Vec2, properties: Properties):
        s = Vec2(start)
        e = Vec2(end)
        next_stage = self.next_stage

        if self.solid_lines_only or len(properties.linetype_pattern) < 2:  # CONTINUOUS
            next_stage.draw_line(s, e, properties)
            return

        renderer = linetypes.LineTypeRenderer(self.pattern(properties))
        next_stage.draw_solid_lines(
            [(s, e) for s, e in renderer.line_segment(s, e)],
            properties,
        )

    def draw_solid_lines(
        self, lines: list[tuple[Vec2, Vec2]], properties: Properties
    ) -> None:
        self.next_stage.draw_solid_lines(lines, properties)

    def draw_path(self, path: BkPath2d, properties: Properties):
        next_stage = self.next_stage

        if self.solid_lines_only or len(properties.linetype_pattern) < 2:  # CONTINUOUS
            next_stage.draw_path(path, properties)
            return

        renderer = linetypes.LineTypeRenderer(self.pattern(properties))
        vertices = path.flattening(self.config.max_flattening_distance, segments=16)
        next_stage.draw_solid_lines(
            [(Vec2(s), Vec2(e)) for s, e in renderer.line_segments(vertices)],
            properties,
        )

    def draw_filled_paths(
        self,
        paths: list[BkPath2d],
        properties: Properties,
    ) -> None:
        self.next_stage.draw_filled_paths(paths, properties)

    def draw_filled_polygon(self, points: BkPoints2d, properties: Properties) -> None:
        self.next_stage.draw_filled_polygon(points, properties)

    def draw_image(self, image_data: ImageData, properties: Properties) -> None:
        self.next_stage.draw_image(image_data, properties)


class BackendStage2d(RenderStage2d):
    """Send data to the output backend."""

    def __init__(
        self,
        backend: BackendInterface,
        converter: Callable[[Properties], BackendProperties],
    ):
        self.backend = backend
        self.converter = converter
        assert not hasattr(self, "next_stage"), "has to be the last render stage"

    def draw_point(self, pos: Vec2, properties: Properties) -> None:
        self.backend.draw_point(pos, self.converter(properties))

    def draw_line(self, start: Vec2, end: Vec2, properties: Properties):
        self.backend.draw_line(start, end, self.converter(properties))

    def draw_solid_lines(
        self, lines: list[tuple[Vec2, Vec2]], properties: Properties
    ) -> None:
        self.backend.draw_solid_lines(lines, self.converter(properties))

    def draw_path(self, path: BkPath2d, properties: Properties):
        self.backend.draw_path(path, self.converter(properties))

    def draw_filled_paths(
        self,
        paths: list[BkPath2d],
        properties: Properties,
    ) -> None:
        self.backend.draw_filled_paths(paths, self.converter(properties))

    def draw_filled_polygon(self, points: BkPoints2d, properties: Properties) -> None:
        self.backend.draw_filled_polygon(points, self.converter(properties))

    def draw_image(self, image_data: ImageData, properties: Properties) -> None:
        self.backend.draw_image(image_data, self.converter(properties))


def _mask_image(image_data: ImageData, outer_bounds: list[BkPoints2d]) -> None:
    """Mask away the clipped parts of the image. The argument `outer_bounds` is only
    used for clip mode "remove_inside". The outer bounds can be composed of multiple
    parts. If `outer_bounds` is empty the image has no removed parts and is fully
    visible before applying the image clipping path.

    Args:
        image_data:
            image_data.pixel_boundary: path contains the image clipping path
            image_data.remove_outside: defines the clipping mode (inside/outside)
        outer_bounds: countain the parts of the image which are __not__ removed by
            clipping through viewports or clipped block references
            e.g. an image without any removed parts has the outer bounds
            [(0, 0) (width, 0), (width, height), (0, height)]

    """
    clip_polygon = [(p.x, p.y) for p in image_data.pixel_boundary_path.vertices()]
    # create an empty image
    clipping_image = PIL.Image.new("L", image_data.image_size(), 0)
    # paint in the clipping path
    PIL.ImageDraw.ImageDraw(clipping_image).polygon(
        clip_polygon, outline=None, width=0, fill=1
    )
    clipping_mask = np.asarray(clipping_image)

    if not image_data.remove_outside:  # clip mode "remove_inside"
        if outer_bounds:
            # create a new empty image
            visible_image = PIL.Image.new("L", image_data.image_size(), 0)
            # paint in parts of the image which are still visible
            for boundary in outer_bounds:
                clip_polygon = [(p.x, p.y) for p in boundary.vertices()]
                PIL.ImageDraw.ImageDraw(visible_image).polygon(
                    clip_polygon, outline=None, width=0, fill=1
                )
            # remove the clipping path
            clipping_mask = np.asarray(visible_image) - clipping_mask
        else:
            # create mask for fully visible image
            fully_visible_image_mask = np.full(
                clipping_mask.shape, fill_value=1, dtype=clipping_mask.dtype
            )
            # remove the clipping path
            clipping_mask = fully_visible_image_mask - clipping_mask
    image_data.image[:, :, 3] *= clipping_mask


def _clip_image_polygon(
    clipping_portal: ClippingPortal, polygon_px: BkPoints2d, m: Matrix44
) -> list[BkPoints2d]:
    original = [polygon_px]

    # inverse matrix includes the transformation applied by the clipping portal
    inverse = clipping_portal.transform_matrix(m)
    try:
        inverse.inverse()
    except ZeroDivisionError:
        # inverse transformation from WCS to pixel coordinates is not possible
        return original

    # transform image coordinates to WCS coordinates
    polygon = polygon_px.clone()
    polygon.transform_inplace(m)

    clipped_polygons = clipping_portal.clip_polygon(polygon)
    if (len(clipped_polygons) == 1) and (clipped_polygons[0] is polygon):
        # this shows the caller that the image boundary path wasn't clipped
        return original
    # transform WCS coordinates to image coordinates
    for polygon in clipped_polygons:
        polygon.transform_inplace(inverse)
    return clipped_polygons  # in image coordinates!


def invert_color(color: Color) -> Color:
    r, g, b = RGB.from_hex(color)
    return RGB(255 - r, 255 - g, 255 - b).to_hex()


def swap_bw(color: str) -> Color:
    color = color.lower()
    if color == "#000000":
        return "#ffffff"
    if color == "#ffffff":
        return "#000000"
    return color


def color_to_monochrome(color: Color, scale: float = 1.0, offset: float = 0.0) -> Color:
    lum = RGB.from_hex(color).luminance * scale + offset
    if lum < 0.0:
        lum = 0.0
    elif lum > 1.0:
        lum = 1.0
    gray = round(lum * 255)
    return RGB(gray, gray, gray).to_hex()


def apply_color_policy(color: Color, policy: ColorPolicy, custom_color: Color) -> Color:
    alpha = color[7:9]
    color = color[:7]
    if policy == ColorPolicy.COLOR_SWAP_BW:
        color = swap_bw(color)
    elif policy == ColorPolicy.COLOR_NEGATIVE:
        color = invert_color(color)
    elif policy == ColorPolicy.MONOCHROME_DARK_BG:  # [0.3, 1.0]
        color = color_to_monochrome(color, scale=0.7, offset=0.3)
    elif policy == ColorPolicy.MONOCHROME_LIGHT_BG:  # [0.0, 0.7]
        color = color_to_monochrome(color, scale=0.7, offset=0.0)
    elif policy == ColorPolicy.MONOCHROME:  # [0.0, 1.0]
        color = color_to_monochrome(color)
    elif policy == ColorPolicy.BLACK:
        color = "#000000"
    elif policy == ColorPolicy.WHITE:
        color = "#ffffff"
    elif policy == ColorPolicy.CUSTOM:
        fg = custom_color
        color = fg[:7]
        alpha = fg[7:9]
    return color + alpha


def filter_vp_entities(
    msp: Layout,
    limits: Sequence[float],
    bbox_cache: Optional[ezdxf.bbox.Cache] = None,
) -> Iterator[DXFGraphic]:
    """Yields all DXF entities that need to be processed by the given viewport
    `limits`. The entities may be partially of even complete outside the viewport.
    By passing the bounding box cache of the modelspace entities,
    the function can filter entities outside the viewport to speed up rendering
    time.

    There are two processing modes for the `bbox_cache`:

        1. The `bbox_cache` is``None``: all entities must be processed,
           pass through mode
        2. If the `bbox_cache` is given but does not contain an entity,
           the bounding box is computed and added to the cache.
           Even passing in an empty cache can speed up rendering time when
           multiple viewports need to be processed.

    Args:
        msp: modelspace layout
        limits: modelspace limits of the viewport, as tuple (min_x, min_y, max_x, max_y)
        bbox_cache: the bounding box cache of the modelspace entities

    """

    # WARNING: this works only with top-view viewports
    # The current state of the drawing add-on supports only top-view viewports!
    def is_visible(e):
        entity_bbox = bbox_cache.get(e)
        if entity_bbox is None:
            # compute and add bounding box
            entity_bbox = ezdxf.bbox.extents((e,), fast=True, cache=bbox_cache)
        if not entity_bbox.has_data:
            return True
        # Check for separating axis:
        if min_x >= entity_bbox.extmax.x:
            return False
        if max_x <= entity_bbox.extmin.x:
            return False
        if min_y >= entity_bbox.extmax.y:
            return False
        if max_y <= entity_bbox.extmin.y:
            return False
        return True

    if bbox_cache is None:  # pass through all entities
        yield from msp
        return

    min_x, min_y, max_x, max_y = limits
    if not bbox_cache.has_data:
        # fill cache at once
        ezdxf.bbox.extents(msp, fast=True, cache=bbox_cache)

    for entity in msp:
        if is_visible(entity):
            yield entity


def prepare_string_for_rendering(text: str, dxftype: str) -> str:
    assert "\n" not in text, "not a single line of text"
    if dxftype in {"TEXT", "ATTRIB", "ATTDEF"}:
        text = replace_non_printable_characters(text, replacement="?")
        text = text.replace("\t", "?")
    elif dxftype == "MTEXT":
        text = replace_non_printable_characters(text, replacement="▯")
        text = text.replace("\t", "        ")
    else:
        raise TypeError(dxftype)
    return text



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import pytest

from ezdxf.lldxf.types import dxftag, render_tag
from ezdxf.math import Vec3


class TestRenderTag:
    def test_render_string_type(self):
        tag = dxftag(1, "text")
        assert render_tag(tag, 0) == "1"
        assert render_tag(tag, 1) == "<str>"
        assert render_tag(tag, 2) == "text"

    def test_render_int_type(self):
        tag = dxftag(70, 1)
        assert render_tag(tag, 0) == "70"
        assert render_tag(tag, 1) == "<int>"
        assert render_tag(tag, 2) == "1"

    def test_render_float_type(self):
        tag = dxftag(40, 1.1)
        assert render_tag(tag, 0) == "40"
        assert render_tag(tag, 1) == "<float>"
        assert render_tag(tag, 2) == "1.1"

    def test_render_handle_type(self):
        tag = dxftag(5, "ABBA")
        assert render_tag(tag, 0) == "5"
        assert render_tag(tag, 1) == "<handle>"
        assert render_tag(tag, 2) == "ABBA"

    def test_render_reference_type(self):
        tag = dxftag(330, "ABBA")
        assert render_tag(tag, 0) == "330"
        assert render_tag(tag, 1) == "<ref>"
        assert render_tag(tag, 2) == "ABBA"

    def test_render_ctrl_type(self):
        tag = dxftag(0, "LINE")
        assert render_tag(tag, 0) == "0"
        assert render_tag(tag, 1) == "<ctrl>"
        assert render_tag(tag, 2) == "LINE"

    def test_render_point_type(self):
        tag = dxftag(10, Vec3(1, 2, 3))
        assert render_tag(tag, 0) == "10"
        assert render_tag(tag, 1) == "<point>"
        assert render_tag(tag, 2) == "(1.0, 2.0, 3.0)"

    def test_render_binary_data_type(self):
        tag = dxftag(310, "FEFE")
        assert render_tag(tag, 0) == "310"
        assert render_tag(tag, 1) == "<bin>"
        assert render_tag(tag, 2) == "FEFE"

    def test_raise_index_error_for_invalid_column_index(self):
        tag = dxftag(1, "text")
        pytest.raises(IndexError, render_tag, tag, 3)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021-2025, Manfred Moitzi
#  License: MIT License

import pytest

import ezdxf
from ezdxf.entities import MText, Text
from ezdxf.layouts import VirtualLayout
from ezdxf.tools.text_size import (
    text_size,
    mtext_size,
    estimate_mtext_extents,
)
from ezdxf.tools.text import (
    set_estimation_safety_factor,
    reset_estimation_safety_factor,
)
from ezdxf.tools.text_layout import leading


@pytest.fixture
def msp():
    yield VirtualLayout()


H3W1 = {"height": 3.0, "width": 1.0}
H2W2 = {"height": 2.0, "width": 2.0}


def test_text_size_of_an_empty_string(msp):
    """The text_size() function does not measure the actual height of a char,
    it always returns the font measurement cap-height for text height and
    cap-height + descender for the total text height.
    """
    text = msp.add_text("", dxfattribs=H3W1)
    size = text_size(text)
    assert size.width == 0.0
    assert size.cap_height == 3.0
    assert size.total_height > 3.3  # assuming the descender factor is > 0.1


def test_text_size_for_height_0():
    text = Text()
    # hack
    text.dxf.__dict__["height"] = 0
    text.dxf.width = 1
    text.dxf.text = "Test"
    size = text_size(text)
    # a text height of 0 should default to 2.5
    assert size.width == 10.0
    assert size.cap_height == 2.5
    assert size.total_height >= 2.5


def test_text_width_of_a_single_char(msp):
    """The "MonospaceFont" font has a char width of cap-height x width factor."""
    text = msp.add_text("X", dxfattribs=H3W1)
    size = text_size(text)
    assert size.width == 3.0, "char width should be equal to cap height"


@pytest.mark.parametrize("s", ["ABC", ".,!", "   "])
def test_text_width_of_a_string(msp, s):
    text = msp.add_text(s, dxfattribs=H3W1)
    size = text_size(text)
    assert size.width == len(s) * size.cap_height


@pytest.mark.parametrize("s", ["ABC", ".,!", "   "])
def test_text_width_of_a_string_for_width_factor_2(msp, s):
    text = msp.add_text(s, dxfattribs=H2W2)
    size = text_size(text)
    assert size.width == len(s) * size.cap_height * 2.0


@pytest.mark.parametrize(
    "s",
    [
        "ABC\n",  # remove line ending
        "ABC\r",  # remove line ending
        "AB^I",  # parse caret notation "^I" -> "\t" (tabulator)
        "AB%%d",  # parse special chars "%%d" -> "°"
    ],
)
def test_measurement_of_plain_text(msp, s):
    text = msp.add_text(s, dxfattribs=H3W1)
    size = text_size(text)
    assert size.width == 3.0 * size.cap_height


def test_support_for_text_size():
    test_string = "TestString"
    doc = ezdxf.new()
    doc.styles.add("OpenSans", font="OpenSans-Regular.ttf")
    text = doc.modelspace().add_text(
        test_string,
        dxfattribs={
            "style": "OpenSans",
            "height": 2.0,
        },
    )
    length = len(test_string)
    size = text_size(text)
    # Do not check exact measurements!
    assert length * 1.0 < size.width < length * 2.0
    assert 1.95 < size.cap_height < 2.05
    assert size.total_height > size.cap_height


def test_mtext_size_of_an_empty_string(msp):
    mtext = msp.add_mtext("", dxfattribs={"char_height": 1.0})
    size = mtext_size(mtext)
    assert size.total_width == 0.0
    assert size.total_height == 0.0
    assert size.column_width == 0.0
    assert size.gutter_width == 0.0
    assert size.column_count == 1
    assert size.column_heights == (0.0,)


def test_mtext_size_for_height_0():
    text = MText()
    # hack
    text.dxf.__dict__["char_height"] = 0
    text.dxf.text = "Test"
    text.dxf.width = 20  # reference column width
    size = mtext_size(text)
    # a char height of 0 should default to 2.5
    assert size.total_height >= 2.5
    assert size.total_width >= 10.0


def test_mtext_size_of_a_single_char(msp):
    set_estimation_safety_factor(1.0)
    # Matplotlib support disabled and using MonospaceFont()
    mtext = msp.add_mtext("X", dxfattribs={"char_height": 2.0})
    size = mtext_size(mtext)
    assert size.total_height == 2.0
    assert size.total_width == pytest.approx(1.8794373744139317)
    assert size.column_width == pytest.approx(1.8794373744139317)
    assert size.gutter_width == 0.0
    assert size.column_count == 1
    reset_estimation_safety_factor()


def test_mtext_size_of_a_string(msp):
    set_estimation_safety_factor(1.0)
    # Matplotlib support disabled and using MonospaceFont()
    mtext = msp.add_mtext("XXX", dxfattribs={"char_height": 2.0})
    size = mtext_size(mtext)
    assert size.total_height == 2.0
    assert size.total_width == pytest.approx(5.6383121232417945)
    assert size.column_width == size.total_width
    assert size.gutter_width == 0.0
    assert size.column_count == 1
    reset_estimation_safety_factor()


def test_estimate_mtext_extents(msp):
    set_estimation_safety_factor(1.0)
    # Matplotlib support disabled and using MonospaceFont()
    mtext = msp.add_mtext(
        "XXXXXXXXXXXX\nYYYY\nZ",  # 5 lines!
        dxfattribs={
            "char_height": 2.0,
            "width": 8.0,
        },
    )
    width, height = estimate_mtext_extents(mtext)
    assert height == pytest.approx(15.336)  # 5 lines!
    assert width == 8.0
    reset_estimation_safety_factor()


@pytest.mark.parametrize(
    "cap_height, expected", [(2.0, 6.703281982585398), (3.0, 10.054922973878098)]
)
def test_mtext_size_of_2_lines(cap_height, expected, msp):
    set_estimation_safety_factor(1.0)
    # Matplotlib support disabled and using MonospaceFont()
    mtext = msp.add_mtext(
        "XXX\nYYYY",
        dxfattribs={
            "char_height": cap_height,
            "line_spacing_factor": 1.0,
        },
    )
    size = mtext_size(mtext)
    expected_total_height = leading(cap_height, line_spacing=1.0) + cap_height
    assert size.total_height == pytest.approx(expected_total_height)
    assert size.total_width == pytest.approx(expected), "expected width of 2nd line"
    assert size.column_width == size.total_width
    reset_estimation_safety_factor()


if __name__ == "__main__":
    pytest.main([__file__])



# cython: language_level=3
# Copyright (c) 2023-2024, Manfred Moitzi
# License: MIT License
from typing_extensions import TypeAlias
import numpy as np
import numpy.typing as npt

import cython
from .vector cimport isclose

cdef extern from "constants.h":
    const double ABS_TOL
    const double REL_TOL


NDArray: TypeAlias = npt.NDArray[np.float64]


def has_clockwise_orientation(vertices: np.ndarray) -> bool:
    """ Returns True if 2D `vertices` have clockwise orientation. Ignores
    z-axis of all vertices.

    Args:
        vertices: numpy array

    Raises:
        ValueError: less than 3 vertices

    """
    if len(vertices) < 3:
        raise ValueError('At least 3 vertices required.')

    return _has_clockwise_orientation(vertices, vertices.shape[0])


@cython.boundscheck(False)
@cython.wraparound(False)
cdef bint _has_clockwise_orientation(double [:, ::1] vertices, Py_ssize_t size):
    cdef Py_ssize_t index
    cdef Py_ssize_t start
    cdef Py_ssize_t last = size - 1
    cdef double s = 0.0
    cdef double p1x = vertices[0][0]
    cdef double p1y = vertices[0][1]
    cdef double p2x = vertices[last][0]
    cdef double p2y = vertices[last][1]

    # Using the same tolerance as the Python implementation:
    cdef bint x_is_close = isclose(p1x, p2x, REL_TOL, ABS_TOL)
    cdef bint y_is_close = isclose(p1y, p2y, REL_TOL, ABS_TOL)

    if x_is_close and y_is_close:
        p1x = vertices[0][0]
        p1y = vertices[0][1]
        start = 1
    else:
        p1x = vertices[last][0]
        p1y = vertices[last][1]
        start = 0

    for index in range(start, size):
        p2x = vertices[index][0]
        p2y = vertices[index][1]
        s += (p2x - p1x) * (p2y + p1y)
        p1x = p2x
        p1y = p2y
    return s > 0.0


def lu_decompose(A: NDArray, m1: int, m2: int) -> tuple[NDArray, NDArray, NDArray]:
    upper: np.ndarray = np.array(A, dtype=np.float64)
    n: int = upper.shape[0]
    lower: np.ndarray = np.zeros((n, m1), dtype=np.float64)

    # Is <np.int_> better to match <int> on all platforms?
    index: np.ndarray = np.zeros((n,), dtype=np.int32)  
    _lu_decompose_cext(upper, lower, index, n, m1, m2)
    return upper, lower, index


@cython.boundscheck(False)
@cython.wraparound(False)
cdef _lu_decompose_cext(
    double [:, ::1] upper, 
    double [:, ::1] lower, 
    int [::1] index, 
    int n, 
    int m1, 
    int m2
):
    cdef int mm = m1 + m2 + 1
    cdef int l = m1
    cdef int i, j, k
    cdef double dum

    for i in range(m1):
        for j in range(m1 - i, mm):
            upper[i][j - l] = upper[i][j]
        l -= 1
        for j in range(mm - l - 1, mm):
            upper[i][j] = 0.0

    l = m1
    for k in range(n):
        dum = upper[k][0]
        i = k
        if l < n:
            l += 1
        for j in range(k + 1, l):
            if abs(upper[j][0]) > abs(dum):
                dum = upper[j][0]
                i = j
        index[k] = i + 1
        if i != k:
            for j in range(mm):
                upper[k][j], upper[i][j] = upper[i][j], upper[k][j]

        for i in range(k + 1, l):
            dum = upper[i][0] / upper[k][0]
            lower[k][i - k - 1] = dum
            for j in range(1, mm):
                upper[i][j - 1] = upper[i][j] - dum * upper[k][j]
            upper[i][mm - 1] = 0.0


def solve_vector_banded_matrix(
    x: NDArray,
    upper: NDArray,
    lower: NDArray,
    index: NDArray,
    m1: int,
    m2: int,
) -> NDArray:
    n: int = upper.shape[0]
    x = np.array(x)  # copy x because array x gets modified
    if x.shape[0] != n:
        raise ValueError(
            "Item count of vector <x> has to match the row count of matrix <upper>."
        )
    _solve_vector_banded_matrix_cext(x, upper, lower, index, n, m1, m2)
    return x


@cython.boundscheck(False)
@cython.wraparound(False)
cdef _solve_vector_banded_matrix_cext(
    double [::1] x,
    double [:, ::1] au,
    double [:, ::1] al,
    int [::1] index,
    int n,
    int m1,
    int m2,
):
    cdef int mm = m1 + m2 + 1
    cdef int l = m1
    cdef int i, j, k
    cdef double dum

    for k in range(n):
        j = index[k] - 1
        if j != k:
            x[k], x[j] = x[j], x[k]
        if l < n:
            l += 1
        for j in range(k + 1, l):
            x[j] -= al[k][j - k - 1] * x[k]

    l = 1
    for i in range(n - 1, -1, -1):
        dum = x[i]
        for k in range(1, l):
            dum -= au[i][k] * x[k + i]
        x[i] = dum / au[i][0]
        if l < mm:
            l += 1
    return x



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import ezdxf


def write_dxf_file(filename):
    doc = ezdxf.new("R12")
    doc.header["$HANDLING"] = 0
    msp = doc.modelspace()
    msp.add_line((0, 0), (1, 0))
    doc.saveas(filename)


def test_read_file_without_handles(tmpdir):
    filename = tmpdir.join("r12_without_handles.dxf")
    write_dxf_file(filename)

    doc = ezdxf.readfile(filename)
    msp = doc.modelspace()
    line = msp.query("LINE").first
    assert line is not None
    assert line.dxf.handle is not None
